#!/usr/bin/env python3
from __future__ import annotations

import argparse
import asyncio
import contextlib
import datetime
import enum
import functools
import io
import os
import re
import subprocess
import sys
import tarfile
import textwrap
import urllib.parse
import zipfile
from collections.abc import Iterator, Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, Any, ClassVar, NamedTuple, TypeVar
from typing_extensions import TypeAlias

import aiohttp
import packaging.specifiers
import packaging.version
import tomli
import tomlkit
from termcolor import colored

ActionLevelSelf = TypeVar("ActionLevelSelf", bound="ActionLevel")


class ActionLevel(enum.IntEnum):
    def __new__(cls: type[ActionLevelSelf], value: int, doc: str) -> ActionLevelSelf:
        member = int.__new__(cls, value)
        member._value_ = value
        member.__doc__ = doc
        return member

    @classmethod
    def from_cmd_arg(cls, cmd_arg: str) -> ActionLevel:
        try:
            return cls[cmd_arg]
        except KeyError:
            raise argparse.ArgumentTypeError(f'Argument must be one of "{list(cls.__members__)}"')

    nothing = 0, "make no changes"
    local = 1, "make changes that affect local repo"
    fork = 2, "make changes that affect remote repo, but won't open PRs against upstream"
    everything = 3, "do everything, e.g. open PRs"


@dataclass
class StubInfo:
    distribution: str
    version_spec: str
    obsolete: bool
    no_longer_updated: bool


def read_typeshed_stub_metadata(stub_path: Path) -> StubInfo:
    with (stub_path / "METADATA.toml").open("rb") as f:
        meta = tomli.load(f)
    return StubInfo(
        distribution=stub_path.name,
        version_spec=meta["version"],
        obsolete="obsolete_since" in meta,
        no_longer_updated=meta.get("no_longer_updated", False),
    )


@dataclass
class PypiReleaseDownload:
    url: str
    packagetype: Annotated[str, "Should hopefully be either 'bdist_wheel' or 'sdist'"]
    filename: str
    version: packaging.version.Version
    upload_date: datetime.datetime


VersionString: TypeAlias = str
ReleaseDownload: TypeAlias = dict[str, Any]


@dataclass
class PypiInfo:
    distribution: str
    pypi_root: str
    releases: dict[VersionString, list[ReleaseDownload]]
    info: dict[str, Any]

    def get_release(self, *, version: VersionString) -> PypiReleaseDownload:
        # prefer wheels, since it's what most users will get / it's pretty easy to mess up MANIFEST
        release_info = sorted(self.releases[version], key=lambda x: bool(x["packagetype"] == "bdist_wheel"))[-1]
        return PypiReleaseDownload(
            url=release_info["url"],
            packagetype=release_info["packagetype"],
            filename=release_info["filename"],
            version=packaging.version.Version(version),
            upload_date=datetime.datetime.fromisoformat(release_info["upload_time"]),
        )

    def get_latest_release(self) -> PypiReleaseDownload:
        return self.get_release(version=self.info["version"])

    def releases_in_descending_order(self) -> Iterator[PypiReleaseDownload]:
        for version in sorted(self.releases, key=packaging.version.Version, reverse=True):
            yield self.get_release(version=version)


async def fetch_pypi_info(distribution: str, session: aiohttp.ClientSession) -> PypiInfo:
    # Cf. # https://warehouse.pypa.io/api-reference/json.html#get--pypi--project_name--json
    pypi_root = f"https://pypi.org/pypi/{urllib.parse.quote(distribution)}"
    async with session.get(f"{pypi_root}/json") as response:
        response.raise_for_status()
        j = await response.json()
        return PypiInfo(distribution=distribution, pypi_root=pypi_root, releases=j["releases"], info=j["info"])


@dataclass
class Update:
    distribution: str
    stub_path: Path
    old_version_spec: str
    new_version_spec: str
    links: dict[str, str]
    diff_analysis: DiffAnalysis | None

    def __str__(self) -> str:
        return f"Updating {self.distribution} from {self.old_version_spec!r} to {self.new_version_spec!r}"


@dataclass
class Obsolete:
    distribution: str
    stub_path: Path
    obsolete_since_version: str
    obsolete_since_date: datetime.datetime
    links: dict[str, str]

    def __str__(self) -> str:
        return f"Marking {self.distribution} as obsolete since {self.obsolete_since_version!r}"


@dataclass
class NoUpdate:
    distribution: str
    reason: str

    def __str__(self) -> str:
        return f"Skipping {self.distribution}: {self.reason}"


async def release_contains_py_typed(release_to_download: PypiReleaseDownload, *, session: aiohttp.ClientSession) -> bool:
    async with session.get(release_to_download.url) as response:
        body = io.BytesIO(await response.read())

    packagetype = release_to_download.packagetype
    if packagetype == "bdist_wheel":
        assert release_to_download.filename.endswith(".whl")
        with zipfile.ZipFile(body) as zf:
            return any(Path(f).name == "py.typed" for f in zf.namelist())
    elif packagetype == "sdist":
        assert release_to_download.filename.endswith(".tar.gz")
        with tarfile.open(fileobj=body, mode="r:gz") as zf:
            return any(Path(f).name == "py.typed" for f in zf.getnames())
    else:
        raise AssertionError(f"Unknown package type: {packagetype!r}")


async def find_first_release_with_py_typed(pypi_info: PypiInfo, *, session: aiohttp.ClientSession) -> PypiReleaseDownload:
    release_iter = pypi_info.releases_in_descending_order()
    while await release_contains_py_typed(release := next(release_iter), session=session):
        first_release_with_py_typed = release
    return first_release_with_py_typed


def _check_spec(updated_spec: str, version: packaging.version.Version) -> str:
    assert version in packaging.specifiers.SpecifierSet(f"=={updated_spec}"), f"{version} not in {updated_spec}"
    return updated_spec


def get_updated_version_spec(spec: str, version: packaging.version.Version) -> str:
    """
    Given the old specifier and an updated version, returns an updated specifier that has the
    specificity of the old specifier, but matches the updated version.

    For example:
    spec="1", version="1.2.3" -> "1.2.3"
    spec="1.0.1", version="1.2.3" -> "1.2.3"
    spec="1.*", version="1.2.3" -> "1.*"
    spec="1.*", version="2.3.4" -> "2.*"
    spec="1.1.*", version="1.2.3" -> "1.2.*"
    spec="1.1.1.*", version="1.2.3" -> "1.2.3.*"
    """
    if not spec.endswith(".*"):
        return _check_spec(version.base_version, version)

    specificity = spec.count(".") if spec.removesuffix(".*") else 0
    rounded_version = version.base_version.split(".")[:specificity]
    rounded_version.extend(["0"] * (specificity - len(rounded_version)))

    return _check_spec(".".join(rounded_version) + ".*", version)


@functools.cache
def get_github_api_headers() -> Mapping[str, str]:
    headers = {"Accept": "application/vnd.github.v3+json"}
    secret = os.environ.get("GITHUB_TOKEN")
    if secret is not None:
        headers["Authorization"] = f"token {secret}" if secret.startswith("ghp") else f"Bearer {secret}"
    return headers


@dataclass
class GithubInfo:
    repo_path: str
    tags: list[dict[str, Any]]


async def get_github_repo_info(session: aiohttp.ClientSession, pypi_info: PypiInfo) -> GithubInfo | None:
    """
    If the project represented by `pypi_info` is hosted on GitHub,
    return information regarding the project as it exists on GitHub.

    Else, return None.
    """
    project_urls = pypi_info.info.get("project_urls", {}).values()
    for project_url in project_urls:
        assert isinstance(project_url, str)
        split_url = urllib.parse.urlsplit(project_url)
        if split_url.netloc == "github.com" and not split_url.query and not split_url.fragment:
            url_path = split_url.path.strip("/")
            if len(Path(url_path).parts) == 2:
                github_tags_info_url = f"https://api.github.com/repos/{url_path}/tags"
                async with session.get(github_tags_info_url, headers=get_github_api_headers()) as response:
                    if response.status == 200:
                        tags = await response.json()
                        assert isinstance(tags, list)
                        return GithubInfo(repo_path=url_path, tags=tags)
    return None


class GithubDiffInfo(NamedTuple):
    repo_path: str
    old_tag: str
    new_tag: str
    diff_url: str


async def get_diff_info(
    session: aiohttp.ClientSession, stub_info: StubInfo, pypi_info: PypiInfo, pypi_version: packaging.version.Version
) -> GithubDiffInfo | None:
    """Return a tuple giving info about the diff between two releases, if possible.

    Return `None` if the project isn't hosted on GitHub,
    or if a link pointing to the diff couldn't be found for any other reason.
    """
    github_info = await get_github_repo_info(session, pypi_info)
    if github_info is None:
        return None

    versions_to_tags = {}
    for tag in github_info.tags:
        tag_name = tag["name"]
        # Some packages in typeshed (e.g. emoji) have tag names
        # that are invalid to be passed to the Version() constructor,
        # e.g. v.1.4.2
        with contextlib.suppress(packaging.version.InvalidVersion):
            versions_to_tags[packaging.version.Version(tag_name)] = tag_name

    curr_specifier = packaging.specifiers.SpecifierSet(f"=={stub_info.version_spec}")

    try:
        new_tag = versions_to_tags[pypi_version]
    except KeyError:
        return None

    try:
        old_version = max(version for version in versions_to_tags if version in curr_specifier)
    except ValueError:
        return None
    else:
        old_tag = versions_to_tags[old_version]

    diff_url = f"https://github.com/{github_info.repo_path}/compare/{old_tag}...{new_tag}"
    async with session.get(diff_url, headers=get_github_api_headers()) as response:
        # Double-check we're returning a valid URL here
        response.raise_for_status()
    return GithubDiffInfo(repo_path=github_info.repo_path, old_tag=old_tag, new_tag=new_tag, diff_url=diff_url)


FileInfo: TypeAlias = dict[str, Any]


def _plural_s(num: int, /) -> str:
    return "s" if num != 1 else ""


@dataclass
class DiffAnalysis:
    MAXIMUM_NUMBER_OF_FILES_TO_LIST: ClassVar[int] = 7
    py_files: list[FileInfo]
    py_files_stubbed_in_typeshed: list[FileInfo]

    @property
    def runtime_definitely_has_consistent_directory_structure_with_typeshed(self) -> bool:
        """
        If 0 .py files in the GitHub diff exist in typeshed's stubs,
        there's a possibility that the .py files might be found
        in a different directory at runtime.

        For example: pyopenssl has its .py files in the `src/OpenSSL/` directory at runtime,
        but in typeshed the stubs are in the `OpenSSL/` directory.
        """
        return bool(self.py_files_stubbed_in_typeshed)

    @functools.cached_property
    def public_files_added(self) -> Sequence[str]:
        return [
            file["filename"]
            for file in self.py_files
            if not re.match("_[^_]", Path(file["filename"]).name) and file["status"] == "added"
        ]

    @functools.cached_property
    def typeshed_files_deleted(self) -> Sequence[str]:
        return [file["filename"] for file in self.py_files_stubbed_in_typeshed if file["status"] == "removed"]

    @functools.cached_property
    def typeshed_files_modified(self) -> Sequence[str]:
        return [file["filename"] for file in self.py_files_stubbed_in_typeshed if file["status"] in {"modified", "renamed"}]

    @property
    def total_lines_added(self) -> int:
        return sum(file["additions"] for file in self.py_files)

    @property
    def total_lines_deleted(self) -> int:
        return sum(file["deletions"] for file in self.py_files)

    def _describe_files(self, *, verb: str, filenames: Sequence[str]) -> str:
        num_files = len(filenames)
        if num_files > 1:
            description = f"have been {verb}"
            # Don't list the filenames if there are *loads* of files
            if num_files <= self.MAXIMUM_NUMBER_OF_FILES_TO_LIST:
                description += ": "
                description += ", ".join(f"`{filename}`" for filename in filenames)
            description += "."
            return description
        if num_files == 1:
            return f"has been {verb}: `{filenames[0]}`."
        return f"have been {verb}."

    def describe_public_files_added(self) -> str:
        num_files_added = len(self.public_files_added)
        analysis = f"{num_files_added} public Python file{_plural_s(num_files_added)} "
        analysis += self._describe_files(verb="added", filenames=self.public_files_added)
        return analysis

    def describe_typeshed_files_deleted(self) -> str:
        num_files_deleted = len(self.typeshed_files_deleted)
        analysis = f"{num_files_deleted} file{_plural_s(num_files_deleted)} included in typeshed's stubs "
        analysis += self._describe_files(verb="deleted", filenames=self.typeshed_files_deleted)
        return analysis

    def describe_typeshed_files_modified(self) -> str:
        num_files_modified = len(self.typeshed_files_modified)
        analysis = f"{num_files_modified} file{_plural_s(num_files_modified)} included in typeshed's stubs "
        analysis += self._describe_files(verb="modified or renamed", filenames=self.typeshed_files_modified)
        return analysis

    def __str__(self) -> str:
        data_points = []
        if self.runtime_definitely_has_consistent_directory_structure_with_typeshed:
            data_points += [
                self.describe_public_files_added(),
                self.describe_typeshed_files_deleted(),
                self.describe_typeshed_files_modified(),
            ]
        data_points += [
            f"Total lines of Python code added: {self.total_lines_added}.",
            f"Total lines of Python code deleted: {self.total_lines_deleted}.",
        ]
        return "Stubsabot analysis of the diff between the two releases:\n - " + "\n - ".join(data_points)


async def analyze_diff(
    github_repo_path: str, stub_path: Path, old_tag: str, new_tag: str, *, session: aiohttp.ClientSession
) -> DiffAnalysis | None:
    url = f"https://api.github.com/repos/{github_repo_path}/compare/{old_tag}...{new_tag}"
    async with session.get(url, headers=get_github_api_headers()) as response:
        response.raise_for_status()
        json_resp = await response.json()
        assert isinstance(json_resp, dict)
    # https://docs.github.com/en/rest/commits/commits#compare-two-commits
    py_files: list[FileInfo] = [file for file in json_resp["files"] if Path(file["filename"]).suffix == ".py"]
    files_in_typeshed = set(stub_path.rglob("*.pyi"))
    py_files_stubbed_in_typeshed = [file for file in py_files if (stub_path / f"{file['filename']}i") in files_in_typeshed]
    return DiffAnalysis(py_files=py_files, py_files_stubbed_in_typeshed=py_files_stubbed_in_typeshed)


async def determine_action(stub_path: Path, session: aiohttp.ClientSession) -> Update | NoUpdate | Obsolete:
    stub_info = read_typeshed_stub_metadata(stub_path)
    if stub_info.obsolete:
        return NoUpdate(stub_info.distribution, "obsolete")
    if stub_info.no_longer_updated:
        return NoUpdate(stub_info.distribution, "no longer updated")

    pypi_info = await fetch_pypi_info(stub_info.distribution, session)
    latest_release = pypi_info.get_latest_release()
    latest_version = latest_release.version
    spec = packaging.specifiers.SpecifierSet(f"=={stub_info.version_spec}")
    if latest_version in spec:
        return NoUpdate(stub_info.distribution, "up to date")

    is_obsolete = await release_contains_py_typed(latest_release, session=session)
    if is_obsolete:
        first_release_with_py_typed = await find_first_release_with_py_typed(pypi_info, session=session)
        relevant_version = version_obsolete_since = first_release_with_py_typed.version
    else:
        relevant_version = latest_version

    project_urls = pypi_info.info["project_urls"] or {}
    maybe_links: dict[str, str | None] = {
        "Release": f"{pypi_info.pypi_root}/{relevant_version}",
        "Homepage": project_urls.get("Homepage"),
        "Changelog": project_urls.get("Changelog") or project_urls.get("Changes") or project_urls.get("Change Log"),
    }
    links = {k: v for k, v in maybe_links.items() if v is not None}

    diff_info = await get_diff_info(session, stub_info, pypi_info, relevant_version)
    if diff_info is not None:
        github_repo_path, old_tag, new_tag, diff_url = diff_info
        links["Diff"] = diff_url

    if is_obsolete:
        return Obsolete(
            stub_info.distribution,
            stub_path,
            obsolete_since_version=str(version_obsolete_since),
            obsolete_since_date=first_release_with_py_typed.upload_date,
            links=links,
        )

    if diff_info is None:
        diff_analysis: DiffAnalysis | None = None
    else:
        diff_analysis = await analyze_diff(
            github_repo_path=github_repo_path, stub_path=stub_path, old_tag=old_tag, new_tag=new_tag, session=session
        )

    return Update(
        distribution=stub_info.distribution,
        stub_path=stub_path,
        old_version_spec=stub_info.version_spec,
        new_version_spec=get_updated_version_spec(stub_info.version_spec, latest_version),
        links=links,
        diff_analysis=diff_analysis,
    )


TYPESHED_OWNER = "python"


@functools.lru_cache()
def get_origin_owner() -> str:
    output = subprocess.check_output(["git", "remote", "get-url", "origin"], text=True).strip()
    match = re.match(r"(git@github.com:|https://github.com/)(?P<owner>[^/]+)/(?P<repo>[^/\s]+)", output)
    assert match is not None, f"Couldn't identify origin's owner: {output!r}"
    assert match.group("repo").removesuffix(".git") == "typeshed", f'Unexpected repo: {match.group("repo")!r}'
    return match.group("owner")


async def create_or_update_pull_request(*, title: str, body: str, branch_name: str, session: aiohttp.ClientSession) -> None:
    fork_owner = get_origin_owner()

    async with session.post(
        f"https://api.github.com/repos/{TYPESHED_OWNER}/typeshed/pulls",
        json={"title": title, "body": body, "head": f"{fork_owner}:{branch_name}", "base": "main"},
        headers=get_github_api_headers(),
    ) as response:
        resp_json = await response.json()
        if response.status == 422 and any(
            "A pull request already exists" in e.get("message", "") for e in resp_json.get("errors", [])
        ):
            # Find the existing PR
            async with session.get(
                f"https://api.github.com/repos/{TYPESHED_OWNER}/typeshed/pulls",
                params={"state": "open", "head": f"{fork_owner}:{branch_name}", "base": "main"},
                headers=get_github_api_headers(),
            ) as response:
                response.raise_for_status()
                resp_json = await response.json()
                assert len(resp_json) >= 1
                pr_number = resp_json[0]["number"]
            # Update the PR's title and body
            async with session.patch(
                f"https://api.github.com/repos/{TYPESHED_OWNER}/typeshed/pulls/{pr_number}",
                json={"title": title, "body": body},
                headers=get_github_api_headers(),
            ) as response:
                response.raise_for_status()
            return
        response.raise_for_status()


def has_non_stubsabot_commits(branch: str) -> bool:
    assert not branch.startswith("origin/")
    try:
        # commits on origin/branch that are not on branch or are
        # patch equivalent to a commit on branch
        print(
            "[debugprint]",
            subprocess.check_output(
                ["git", "log", "--right-only", "--pretty=%an %s", "--cherry-pick", f"{branch}...origin/{branch}"]
            ),
        )
        print(
            "[debugprint]",
            subprocess.check_output(
                ["git", "log", "--right-only", "--pretty=%an", "--cherry-pick", f"{branch}...origin/{branch}"]
            ),
        )
        output = subprocess.check_output(
            ["git", "log", "--right-only", "--pretty=%an", "--cherry-pick", f"{branch}...origin/{branch}"],
            stderr=subprocess.DEVNULL,
        )
        return bool(set(output.splitlines()) - {b"stubsabot"})
    except subprocess.CalledProcessError:
        # origin/branch does not exist
        return False


def latest_commit_is_different_to_last_commit_on_origin(branch: str) -> bool:
    assert not branch.startswith("origin/")
    try:
        # https://www.git-scm.com/docs/git-range-diff
        # If the number of lines is >1,
        # it indicates that something about our commit is different to the last commit
        # (Could be the commit "content", or the commit message).
        commit_comparison = subprocess.run(
            ["git", "range-diff", f"origin/{branch}~1..origin/{branch}", "HEAD~1..HEAD"], check=True, capture_output=True
        )
        return len(commit_comparison.stdout.splitlines()) > 1
    except subprocess.CalledProcessError:
        # origin/branch does not exist
        return True


class RemoteConflict(Exception):
    pass


def somewhat_safe_force_push(branch: str) -> None:
    if has_non_stubsabot_commits(branch):
        raise RemoteConflict(f"origin/{branch} has non-stubsabot changes that are not on {branch}!")
    subprocess.check_call(["git", "push", "origin", branch, "--force"])


def normalize(name: str) -> str:
    # PEP 503 normalization
    return re.sub(r"[-_.]+", "-", name).lower()


# lock should be unnecessary, but can't hurt to enforce mutual exclusion
_repo_lock = asyncio.Lock()

BRANCH_PREFIX = "stubsabot"


def get_update_pr_body(update: Update, metadata: dict[str, Any]) -> str:
    body = "\n".join(f"{k}: {v}" for k, v in update.links.items())

    if update.diff_analysis is not None:
        body += f"\n\n{update.diff_analysis}"

    stubtest_will_run = not metadata.get("stubtest", {}).get("skip", False)
    if stubtest_will_run:
        body += textwrap.dedent(
            """

            If stubtest fails for this PR:
            - Leave this PR open (as a reminder, and to prevent stubsabot from opening another PR)
            - Fix stubtest failures in another PR, then close this PR

            Note that you will need to close and re-open the PR in order to trigger CI
            """
        )
    else:
        body += textwrap.dedent(
            f"""

            :warning: Review this PR manually, as stubtest is skipped in CI for {update.distribution}! :warning:
            """
        )
    return body


async def suggest_typeshed_update(update: Update, session: aiohttp.ClientSession, action_level: ActionLevel) -> None:
    if action_level <= ActionLevel.nothing:
        return
    title = f"[stubsabot] Bump {update.distribution} to {update.new_version_spec}"
    async with _repo_lock:
        branch_name = f"{BRANCH_PREFIX}/{normalize(update.distribution)}"
        subprocess.check_call(["git", "checkout", "-B", branch_name, "origin/main"])
        with open(update.stub_path / "METADATA.toml", "rb") as f:
            meta = tomlkit.load(f)
        meta["version"] = update.new_version_spec
        with open(update.stub_path / "METADATA.toml", "w", encoding="UTF-8") as f:
            tomlkit.dump(meta, f)
        body = get_update_pr_body(update, meta)
        subprocess.check_call(["git", "commit", "--all", "-m", f"{title}\n\n{body}"])
        if action_level <= ActionLevel.local:
            return
        if not latest_commit_is_different_to_last_commit_on_origin(branch_name):
            print(f"No pushing to origin required: origin/{branch_name} exists and requires no changes!")
            return
        somewhat_safe_force_push(branch_name)
        if action_level <= ActionLevel.fork:
            return

    await create_or_update_pull_request(title=title, body=body, branch_name=branch_name, session=session)


async def suggest_typeshed_obsolete(obsolete: Obsolete, session: aiohttp.ClientSession, action_level: ActionLevel) -> None:
    if action_level <= ActionLevel.nothing:
        return
    title = f"[stubsabot] Mark {obsolete.distribution} as obsolete since {obsolete.obsolete_since_version}"
    async with _repo_lock:
        branch_name = f"{BRANCH_PREFIX}/{normalize(obsolete.distribution)}"
        subprocess.check_call(["git", "checkout", "-B", branch_name, "origin/main"])
        with open(obsolete.stub_path / "METADATA.toml", "rb") as f:
            meta = tomlkit.load(f)
        obs_string = tomlkit.string(obsolete.obsolete_since_version)
        obs_string.comment(f"Released on {obsolete.obsolete_since_date.date().isoformat()}")
        meta["obsolete_since"] = obs_string
        with open(obsolete.stub_path / "METADATA.toml", "w", encoding="UTF-8") as f:
            tomlkit.dump(meta, f)
        body = "\n".join(f"{k}: {v}" for k, v in obsolete.links.items())
        subprocess.check_call(["git", "commit", "--all", "-m", f"{title}\n\n{body}"])
        if action_level <= ActionLevel.local:
            return
        if not latest_commit_is_different_to_last_commit_on_origin(branch_name):
            print(f"No PR required: origin/{branch_name} exists and requires no changes!")
            return
        somewhat_safe_force_push(branch_name)
        if action_level <= ActionLevel.fork:
            return

    await create_or_update_pull_request(title=title, body=body, branch_name=branch_name, session=session)


async def main() -> None:
    assert sys.version_info >= (3, 9)

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--action-level",
        type=ActionLevel.from_cmd_arg,
        default=ActionLevel.everything,
        help="Limit actions performed to achieve dry runs for different levels of dryness",
    )
    parser.add_argument(
        "--action-count-limit",
        type=int,
        default=None,
        help="Limit number of actions performed and the remainder are logged. Useful for testing",
    )
    args = parser.parse_args()

    if args.action_level > ActionLevel.nothing:
        subprocess.run(["git", "update-index", "--refresh"], capture_output=True)
        diff_result = subprocess.run(["git", "diff-index", "HEAD", "--name-only"], text=True, capture_output=True)
        if diff_result.returncode:
            print("Unexpected exception!")
            print(diff_result.stdout)
            print(diff_result.stderr)
            sys.exit(diff_result.returncode)
        if diff_result.stdout:
            changed_files = ", ".join(repr(line) for line in diff_result.stdout.split("\n") if line)
            print(f"Cannot run stubsabot, as uncommitted changes are present in {changed_files}!")
            sys.exit(1)

    if args.action_level > ActionLevel.fork:
        if os.environ.get("GITHUB_TOKEN") is None:
            raise ValueError("GITHUB_TOKEN environment variable must be set")

    denylist = {"gdb"}  # gdb is not a pypi distribution

    original_branch = subprocess.run(
        ["git", "branch", "--show-current"], text=True, capture_output=True, check=True
    ).stdout.strip()

    if args.action_level >= ActionLevel.fork:
        subprocess.check_call(["git", "fetch", "--prune", "--all"])

    try:
        conn = aiohttp.TCPConnector(limit_per_host=10)
        async with aiohttp.ClientSession(connector=conn) as session:
            tasks = [
                asyncio.create_task(determine_action(stubs_path, session))
                for stubs_path in Path("stubs").iterdir()
                if stubs_path.name not in denylist
            ]

            action_count = 0
            for task in asyncio.as_completed(tasks):
                update = await task
                print(update)

                if isinstance(update, NoUpdate):
                    continue

                if args.action_count_limit is not None and action_count >= args.action_count_limit:
                    print(colored("... but we've reached action count limit", "red"))
                    continue
                action_count += 1

                try:
                    if isinstance(update, Update):
                        await suggest_typeshed_update(update, session, action_level=args.action_level)
                        continue
                    if isinstance(update, Obsolete):
                        await suggest_typeshed_obsolete(update, session, action_level=args.action_level)
                        continue
                except RemoteConflict as e:
                    print(colored(f"... but ran into {type(e).__qualname__}: {e}", "red"))
                    continue
                raise AssertionError
    finally:
        # if you need to cleanup, try:
        # git branch -D $(git branch --list 'stubsabot/*')
        if args.action_level >= ActionLevel.local:
            subprocess.check_call(["git", "checkout", original_branch])


if __name__ == "__main__":
    asyncio.run(main())



from typing import Any

EOF: Any
E: Any
namespaces: Any
scopingElements: Any
formattingElements: Any
specialElements: Any
htmlIntegrationPointElements: Any
mathmlTextIntegrationPointElements: Any
adjustSVGAttributes: Any
adjustMathMLAttributes: Any
adjustForeignAttributes: Any
unadjustForeignAttributes: Any
spaceCharacters: Any
tableInsertModeElements: Any
asciiLowercase: Any
asciiUppercase: Any
asciiLetters: Any
digits: Any
hexDigits: Any
asciiUpper2Lower: Any
headingElements: Any
voidElements: Any
cdataElements: Any
rcdataElements: Any
booleanAttributes: Any
entitiesWindows1252: Any
xmlEntities: Any
entities: Any
replacementCharacters: Any
tokenTypes: Any
tagTokenTypes: Any
prefixes: Any

class DataLossWarning(UserWarning): ...
class _ReparseException(Exception): ...



"""Utilities that are imported by multiple scripts in the tests directory."""

from __future__ import annotations

import os
import re
from functools import cache
from itertools import filterfalse
from pathlib import Path
from typing import NamedTuple

import pathspec  # type: ignore[import]
import tomli


def strip_comments(text: str) -> str:
    return text.split("#")[0].strip()


try:
    from termcolor import colored as colored
except ImportError:

    def colored(s: str, _: str) -> str:  # type: ignore[misc]
        return s


def print_error(error: str, end: str = "\n", fix_path: tuple[str, str] = ("", "")) -> None:
    error_split = error.split("\n")
    old, new = fix_path
    for line in error_split[:-1]:
        print(colored(line.replace(old, new), "red"))
    print(colored(error_split[-1], "red"), end=end)


def print_success_msg() -> None:
    print(colored("success", "green"))


# ====================================================================
# Reading dependencies from METADATA.toml files
# ====================================================================


@cache
def read_dependencies(distribution: str) -> tuple[str, ...]:
    with Path("stubs", distribution, "METADATA.toml").open("rb") as f:
        data = tomli.load(f)
    requires = data.get("requires", [])
    assert isinstance(requires, list)
    dependencies = []
    for dependency in requires:
        assert isinstance(dependency, str)
        assert dependency.startswith("types-"), f"unrecognized dependency {dependency!r}"
        dependencies.append(dependency[6:].split("<")[0])
    return tuple(dependencies)


def get_recursive_requirements(package_name: str, seen: set[str] | None = None) -> list[str]:
    seen = seen if seen is not None else {package_name}
    for dependency in filterfalse(seen.__contains__, read_dependencies(package_name)):
        seen.update(get_recursive_requirements(dependency, seen))
    return sorted(seen | {package_name})


# ====================================================================
# Parsing the stdlib/VERSIONS file
# ====================================================================


VERSIONS_RE = re.compile(r"^([a-zA-Z_][a-zA-Z0-9_.]*): ([23]\.\d{1,2})-([23]\.\d{1,2})?$")


# ====================================================================
# Getting test-case directories from package names
# ====================================================================


class PackageInfo(NamedTuple):
    name: str
    test_case_directory: Path


def testcase_dir_from_package_name(package_name: str) -> Path:
    return Path("stubs", package_name, "@tests/test_cases")


def get_all_testcase_directories() -> list[PackageInfo]:
    testcase_directories = [PackageInfo("stdlib", Path("test_cases"))]
    for package_name in os.listdir("stubs"):
        potential_testcase_dir = testcase_dir_from_package_name(package_name)
        if potential_testcase_dir.is_dir():
            testcase_directories.append(PackageInfo(package_name, potential_testcase_dir))
    return sorted(testcase_directories)


# ====================================================================
# Parsing .gitignore
# ====================================================================


@cache
def get_gitignore_spec() -> pathspec.PathSpec:
    with open(".gitignore", encoding="UTF-8") as f:
        return pathspec.PathSpec.from_lines("gitwildmatch", f.readlines())


def spec_matches_path(spec: pathspec.PathSpec, path: Path) -> bool:
    normalized_path = path.as_posix()
    if path.is_dir():
        normalized_path += "/"
    return spec.match_file(normalized_path)  # type: ignore[no-any-return]



import abc
from collections.abc import ItemsView, KeysView, Mapping, Sequence, ValuesView
from typing import Any

from ..cresultproxy import BaseRow as BaseRow

MD_INDEX: int

def rowproxy_reconstructor(cls, state): ...

KEY_INTEGER_ONLY: int
KEY_OBJECTS_ONLY: int
KEY_OBJECTS_BUT_WARN: int
KEY_OBJECTS_NO_WARN: int

class Row(BaseRow, Sequence[Any], metaclass=abc.ABCMeta):
    @property
    def count(self): ...
    @property
    def index(self): ...
    def __contains__(self, key): ...
    __hash__ = BaseRow.__hash__
    def __lt__(self, other): ...
    def __le__(self, other): ...
    def __ge__(self, other): ...
    def __gt__(self, other): ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...
    def keys(self): ...

class LegacyRow(Row, metaclass=abc.ABCMeta):
    def __contains__(self, key): ...
    def has_key(self, key): ...
    def items(self): ...
    def iterkeys(self): ...
    def itervalues(self): ...
    def values(self): ...

BaseRowProxy = BaseRow
RowProxy = Row

class ROMappingView(KeysView[Any], ValuesView[Any], ItemsView[Any, Any]):
    def __init__(self, mapping, items) -> None: ...
    def __len__(self): ...
    def __iter__(self): ...
    def __contains__(self, item): ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...

class RowMapping(BaseRow, Mapping[Any, Any]):
    __getitem__: Any
    def __iter__(self): ...
    def __len__(self): ...
    def __contains__(self, key): ...
    def items(self): ...
    def keys(self): ...
    def values(self): ...



import io
from _typeshed import Incomplete
from typing_extensions import TypeAlias

NativeIO: TypeAlias = io.StringIO

class Verifier:
    ffi: Incomplete
    preamble: Incomplete
    flags: Incomplete
    kwds: Incomplete
    tmpdir: Incomplete
    sourcefilename: Incomplete
    modulefilename: Incomplete
    ext_package: Incomplete
    def __init__(
        self,
        ffi,
        preamble,
        tmpdir: Incomplete | None = ...,
        modulename: Incomplete | None = ...,
        ext_package: Incomplete | None = ...,
        tag: str = ...,
        force_generic_engine: bool = ...,
        source_extension: str = ...,
        flags: Incomplete | None = ...,
        relative_to: Incomplete | None = ...,
        **kwds,
    ) -> None: ...
    def write_source(self, file: Incomplete | None = ...) -> None: ...
    def compile_module(self) -> None: ...
    def load_library(self): ...
    def get_module_name(self): ...
    def get_extension(self): ...
    def generates_python_module(self): ...
    def make_relative_to(self, kwds, relative_to): ...

def set_tmpdir(dirname) -> None: ...
def cleanup_tmpdir(tmpdir: Incomplete | None = ..., keep_so: bool = ...) -> None: ...



import abc
import sys
from _collections_abc import dict_items, dict_keys, dict_values
from _typeshed import IdentityFunction, Self
from collections.abc import Mapping
from typing import Any, ClassVar, Generic, TypeVar, overload, type_check_only
from typing_extensions import Never

_T = TypeVar("_T")
_U = TypeVar("_U")

# Internal mypy fallback type for all typed dicts (does not exist at runtime)
# N.B. Keep this mostly in sync with typing(_extensions)._TypedDict
@type_check_only
class _TypedDict(Mapping[str, object], metaclass=abc.ABCMeta):
    __total__: ClassVar[bool]
    # Unlike typing(_extensions).TypedDict,
    # subclasses of mypy_extensions.TypedDict do NOT have the __required_keys__ and __optional_keys__ ClassVars
    def copy(self: Self) -> Self: ...
    # Using Never so that only calls using mypy plugin hook that specialize the signature
    # can go through.
    def setdefault(self, k: Never, default: object) -> object: ...
    # Mypy plugin hook for 'pop' expects that 'default' has a type variable type.
    def pop(self, k: Never, default: _T = ...) -> object: ...  # pyright: ignore[reportInvalidTypeVarUse]
    def update(self: Self, __m: Self) -> None: ...
    def items(self) -> dict_items[str, object]: ...
    def keys(self) -> dict_keys[str, object]: ...
    def values(self) -> dict_values[str, object]: ...
    def __delitem__(self, k: Never) -> None: ...
    if sys.version_info >= (3, 9):
        def __or__(self: Self, __other: Self) -> Self: ...
        def __ior__(self: Self, __other: Self) -> Self: ...

def TypedDict(typename: str, fields: dict[str, type[Any]], total: bool = ...) -> type[dict[str, Any]]: ...
@overload
def Arg(type: _T, name: str | None = ...) -> _T: ...
@overload
def Arg(*, name: str | None = ...) -> Any: ...
@overload
def DefaultArg(type: _T, name: str | None = ...) -> _T: ...
@overload
def DefaultArg(*, name: str | None = ...) -> Any: ...
@overload
def NamedArg(type: _T, name: str | None = ...) -> _T: ...
@overload
def NamedArg(*, name: str | None = ...) -> Any: ...
@overload
def DefaultNamedArg(type: _T, name: str | None = ...) -> _T: ...
@overload
def DefaultNamedArg(*, name: str | None = ...) -> Any: ...
@overload
def VarArg(type: _T) -> _T: ...
@overload
def VarArg() -> Any: ...
@overload
def KwArg(type: _T) -> _T: ...
@overload
def KwArg() -> Any: ...

# Return type that indicates a function does not return.
# Deprecated: Use typing.NoReturn instead.
class NoReturn: ...

# This is consistent with implementation. Usage intends for this as
# a class decorator, but mypy does not support type[_T] for abstract
# classes until this issue is resolved, https://github.com/python/mypy/issues/4717.
def trait(cls: _T) -> _T: ...
def mypyc_attr(*attrs: str, **kwattrs: object) -> IdentityFunction: ...

class FlexibleAlias(Generic[_T, _U]): ...



from _typeshed import Self

from stripe.api_resources.abstract.api_resource import APIResource as APIResource

class DeletableAPIResource(APIResource):
    @classmethod
    def delete(cls: type[Self], sid: str = ..., **params) -> Self: ...



import sys
from collections.abc import Iterable
from typing import NamedTuple

__all__ = ["RobotFileParser"]

class RequestRate(NamedTuple):
    requests: int
    seconds: int

class RobotFileParser:
    def __init__(self, url: str = ...) -> None: ...
    def set_url(self, url: str) -> None: ...
    def read(self) -> None: ...
    def parse(self, lines: Iterable[str]) -> None: ...
    def can_fetch(self, useragent: str, url: str) -> bool: ...
    def mtime(self) -> int: ...
    def modified(self) -> None: ...
    def crawl_delay(self, useragent: str) -> str | None: ...
    def request_rate(self, useragent: str) -> RequestRate | None: ...
    if sys.version_info >= (3, 8):
        def site_maps(self) -> list[str] | None: ...



from typing import Any

class OAuth2Token(dict[Any, Any]):
    def __init__(self, params, old_scope: Any | None = ...) -> None: ...
    @property
    def scope_changed(self): ...
    @property
    def old_scope(self): ...
    @property
    def old_scopes(self): ...
    @property
    def scope(self): ...
    @property
    def scopes(self): ...
    @property
    def missing_scopes(self): ...
    @property
    def additional_scopes(self): ...

def prepare_mac_header(
    token,
    uri,
    key,
    http_method,
    nonce: Any | None = ...,
    headers: Any | None = ...,
    body: Any | None = ...,
    ext: str = ...,
    hash_algorithm: str = ...,
    issue_time: Any | None = ...,
    draft: int = ...,
): ...
def prepare_bearer_uri(token, uri): ...
def prepare_bearer_headers(token, headers: Any | None = ...): ...
def prepare_bearer_body(token, body: str = ...): ...
def random_token_generator(request, refresh_token: bool = ...): ...
def signed_token_generator(private_pem, **kwargs): ...
def get_token_from_header(request): ...

class TokenBase:
    def __call__(self, request, refresh_token: bool = ...) -> None: ...
    def validate_request(self, request) -> None: ...
    def estimate_type(self, request) -> None: ...

class BearerToken(TokenBase):
    request_validator: Any
    token_generator: Any
    refresh_token_generator: Any
    expires_in: Any
    def __init__(
        self,
        request_validator: Any | None = ...,
        token_generator: Any | None = ...,
        expires_in: Any | None = ...,
        refresh_token_generator: Any | None = ...,
    ) -> None: ...
    def create_token(self, request, refresh_token: bool = ..., **kwargs): ...
    def validate_request(self, request): ...
    def estimate_type(self, request): ...



from .. import event

class PoolEvents(event.Events):
    def connect(self, dbapi_connection, connection_record) -> None: ...
    def first_connect(self, dbapi_connection, connection_record) -> None: ...
    def checkout(self, dbapi_connection, connection_record, connection_proxy) -> None: ...
    def checkin(self, dbapi_connection, connection_record) -> None: ...
    def reset(self, dbapi_connection, connection_record) -> None: ...
    def invalidate(self, dbapi_connection, connection_record, exception) -> None: ...
    def soft_invalidate(self, dbapi_connection, connection_record, exception) -> None: ...
    def close(self, dbapi_connection, connection_record) -> None: ...
    def detach(self, dbapi_connection, connection_record) -> None: ...
    def close_detached(self, dbapi_connection) -> None: ...



from typing import Any

from braintree import exceptions as exceptions
from braintree.configuration import Configuration as Configuration
from braintree.signature_service import SignatureService as SignatureService
from braintree.util.crypto import Crypto as Crypto

class ClientToken:
    @staticmethod
    def generate(params: Any | None = ..., gateway: Any | None = ...): ...
    @staticmethod
    def generate_signature(): ...



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class DifferentialStyle(Serialisable):
    tagname: str
    __elements__: Any
    font: Any
    numFmt: Any
    fill: Any
    alignment: Any
    border: Any
    protection: Any
    extLst: Any
    def __init__(
        self,
        font: Any | None = ...,
        numFmt: Any | None = ...,
        fill: Any | None = ...,
        alignment: Any | None = ...,
        border: Any | None = ...,
        protection: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...

class DifferentialStyleList(Serialisable):
    tagname: str
    dxf: Any
    styles: Any
    __attrs__: Any
    def __init__(self, dxf=..., count: Any | None = ...) -> None: ...
    def append(self, dxf) -> None: ...
    def add(self, dxf): ...
    def __bool__(self): ...
    def __getitem__(self, idx): ...
    @property
    def count(self): ...



from typing import Any

from .base import GrantTypeBase as GrantTypeBase

log: Any

class ImplicitGrant(GrantTypeBase):
    response_types: Any
    grant_allows_refresh_token: bool
    def create_authorization_response(self, request, token_handler): ...
    def create_token_response(self, request, token_handler): ...
    def validate_authorization_request(self, request): ...
    def validate_token_request(self, request): ...



from typing import Any

MIN_ROW: int
MIN_COLUMN: int
MAX_COLUMN: int
MAX_ROW: int
PACKAGE_PROPS: str
PACKAGE_XL: str
PACKAGE_RELS: str
PACKAGE_THEME: Any
PACKAGE_WORKSHEETS: Any
PACKAGE_CHARTSHEETS: Any
PACKAGE_DRAWINGS: Any
PACKAGE_CHARTS: Any
PACKAGE_IMAGES: Any
PACKAGE_WORKSHEET_RELS: Any
PACKAGE_CHARTSHEETS_RELS: Any
PACKAGE_PIVOT_TABLE: Any
PACKAGE_PIVOT_CACHE: Any
ARC_CONTENT_TYPES: str
ARC_ROOT_RELS: Any
ARC_WORKBOOK_RELS: Any
ARC_CORE: Any
ARC_APP: Any
ARC_WORKBOOK: Any
ARC_STYLE: Any
ARC_THEME: Any
ARC_SHARED_STRINGS: Any
ARC_CUSTOM_UI: str
XML_NS: str
DCORE_NS: str
DCTERMS_NS: str
DCTERMS_PREFIX: str
DOC_NS: str
REL_NS: Any
COMMENTS_NS: Any
IMAGE_NS: Any
VML_NS: Any
VTYPES_NS: Any
XPROPS_NS: Any
EXTERNAL_LINK_NS: Any
PKG_NS: str
PKG_REL_NS: Any
COREPROPS_NS: Any
CONTYPES_NS: Any
XSI_NS: str
SHEET_MAIN_NS: str
CHART_NS: str
DRAWING_NS: str
SHEET_DRAWING_NS: str
CHART_DRAWING_NS: str
CUSTOMUI_NS: str
NAMESPACES: Any
WORKBOOK_MACRO: str
WORKBOOK: str
SPREADSHEET: str
SHARED_STRINGS: Any
EXTERNAL_LINK: Any
WORKSHEET_TYPE: Any
COMMENTS_TYPE: Any
STYLES_TYPE: Any
CHARTSHEET_TYPE: Any
DRAWING_TYPE: str
CHART_TYPE: str
CHARTSHAPE_TYPE: str
THEME_TYPE: str
XLTM: Any
XLSM: Any
XLTX: Any
XLSX: Any
EXT_TYPES: Any
CTRL: str
ACTIVEX: str
VBA: str



from distutils.core import Extension as _Extension
from typing import Any

have_pyrex: Any

class Extension(_Extension):
    py_limited_api: Any
    def __init__(self, name, sources, *args, **kw) -> None: ...

class Library(Extension): ...



from typing import Any

log: Any
FOLLOWER_IDENT: Any

class register:
    fns: Any
    @classmethod
    def init(cls, fn): ...
    def for_db(self, *dbnames): ...
    def __call__(self, cfg, *arg): ...

def create_follower_db(follower_ident) -> None: ...
def setup_config(db_url, options, file_config, follower_ident): ...
def drop_follower_db(follower_ident) -> None: ...
def generate_db_urls(db_urls, extra_drivers) -> None: ...
def generate_driver_url(url, driver, query_str): ...
def drop_all_schema_objects_pre_tables(cfg, eng) -> None: ...
def drop_all_schema_objects_post_tables(cfg, eng) -> None: ...
def drop_all_schema_objects(cfg, eng) -> None: ...
def create_db(cfg, eng, ident) -> None: ...
def drop_db(cfg, eng, ident) -> None: ...
def update_db_opts(cfg, db_opts) -> None: ...
def post_configure_engine(url, engine, follower_ident) -> None: ...
def follower_url_from_main(url, ident): ...
def configure_follower(cfg, ident) -> None: ...
def run_reap_dbs(url, ident) -> None: ...
def reap_dbs(idents_file) -> None: ...
def temp_table_keyword_args(cfg, eng) -> None: ...
def prepare_for_drop_tables(config, connection) -> None: ...
def stop_test_class_outside_fixtures(config, db, testcls) -> None: ...
def get_temp_table_name(cfg, eng, base_name): ...
def set_default_schema_on_connection(cfg, dbapi_connection, schema_name) -> None: ...



from typing import Any, Generic, TypeVar
from typing_extensions import Literal

from .. import util
from ..util import HasMemoized, memoized_property
from . import operators, roles
from .annotation import Annotated, SupportsWrappingAnnotations
from .base import Executable, Immutable, SingletonConstant
from .traversals import HasCopyInternals, MemoizedHasCacheKey
from .visitors import Traversible

_T = TypeVar("_T")

def collate(expression, collation): ...
def between(expr, lower_bound, upper_bound, symmetric: bool = ...): ...
def literal(value, type_: Any | None = ...): ...
def outparam(key, type_: Any | None = ...): ...
def not_(clause): ...

class ClauseElement(roles.SQLRole, SupportsWrappingAnnotations, MemoizedHasCacheKey, HasCopyInternals, Traversible):
    __visit_name__: str
    supports_execution: bool
    stringify_dialect: str
    bind: Any
    description: Any
    is_clause_element: bool
    is_selectable: bool
    @property
    def entity_namespace(self) -> None: ...
    def unique_params(self, *optionaldict, **kwargs): ...
    def params(self, *optionaldict, **kwargs): ...
    def compare(self, other, **kw): ...
    def self_group(self, against: Any | None = ...): ...
    def compile(self, bind: Any | None = ..., dialect: Any | None = ..., **kw): ...
    def __invert__(self): ...
    def __bool__(self) -> None: ...
    __nonzero__: Any

class ColumnElement(
    roles.ColumnArgumentOrKeyRole,
    roles.StatementOptionRole,
    roles.WhereHavingRole,
    roles.BinaryElementRole,
    roles.OrderByRole,
    roles.ColumnsClauseRole,
    roles.LimitOffsetRole,
    roles.DMLColumnRole,
    roles.DDLConstraintColumnRole,
    roles.DDLExpressionRole,
    operators.ColumnOperators[_T],
    ClauseElement,
    Generic[_T],
):
    __visit_name__: str
    primary_key: bool
    foreign_keys: Any
    key: Any
    def self_group(self, against: Any | None = ...): ...
    @memoized_property
    def type(self): ...
    @HasMemoized.memoized_attribute
    def comparator(self): ...
    def __getattr__(self, key): ...
    def operate(self, op, *other, **kwargs): ...
    def reverse_operate(self, op, other, **kwargs): ...
    @property
    def expression(self): ...
    @memoized_property
    def base_columns(self): ...
    @memoized_property
    def proxy_set(self): ...
    def shares_lineage(self, othercolumn): ...
    def cast(self, type_): ...
    def label(self, name): ...
    @property
    def anon_label(self): ...
    @property
    def anon_key_label(self): ...

class WrapsColumnExpression:
    @property
    def wrapped_column_expression(self) -> None: ...

class BindParameter(roles.InElementRole, ColumnElement[_T], Generic[_T]):
    __visit_name__: str
    inherit_cache: bool
    key: Any
    unique: Any
    value: Any
    callable: Any
    isoutparam: Any
    required: Any
    expanding: Any
    expand_op: Any
    literal_execute: Any
    type: Any
    def __init__(
        self,
        key,
        value=...,
        type_: Any | None = ...,
        unique: bool = ...,
        required=...,
        quote: Any | None = ...,
        callable_: Any | None = ...,
        expanding: bool = ...,
        isoutparam: bool = ...,
        literal_execute: bool = ...,
        _compared_to_operator: Any | None = ...,
        _compared_to_type: Any | None = ...,
        _is_crud: bool = ...,
    ) -> None: ...
    @property
    def effective_value(self): ...
    def render_literal_execute(self): ...

class TypeClause(ClauseElement):
    __visit_name__: str
    type: Any
    def __init__(self, type_) -> None: ...

class TextClause(
    roles.DDLConstraintColumnRole,
    roles.DDLExpressionRole,
    roles.StatementOptionRole,
    roles.WhereHavingRole,
    roles.OrderByRole,
    roles.FromClauseRole,
    roles.SelectStatementRole,
    roles.BinaryElementRole,
    roles.InElementRole,
    Executable,
    ClauseElement,
):
    __visit_name__: str
    def __and__(self, other): ...
    key: Any
    text: Any
    def __init__(self, text, bind: Any | None = ...): ...
    def bindparams(self, *binds, **names_to_values) -> None: ...
    def columns(self, *cols, **types): ...
    @property
    def type(self): ...
    @property
    def comparator(self): ...
    def self_group(self, against: Any | None = ...): ...

class Null(SingletonConstant, roles.ConstExprRole, ColumnElement[None]):
    __visit_name__: str
    @memoized_property
    def type(self): ...

class False_(SingletonConstant, roles.ConstExprRole, ColumnElement[Literal[False]]):
    __visit_name__: str
    @memoized_property
    def type(self): ...

class True_(SingletonConstant, roles.ConstExprRole, ColumnElement[Literal[True]]):
    __visit_name__: str
    @memoized_property
    def type(self): ...

class ClauseList(roles.InElementRole, roles.OrderByRole, roles.ColumnsClauseRole, roles.DMLColumnRole, ClauseElement):
    __visit_name__: str
    operator: Any
    group: Any
    group_contents: Any
    clauses: Any
    def __init__(self, *clauses, **kwargs) -> None: ...
    def __iter__(self): ...
    def __len__(self): ...
    def append(self, clause) -> None: ...
    def self_group(self, against: Any | None = ...): ...

class BooleanClauseList(ClauseList, ColumnElement[Any]):
    __visit_name__: str
    inherit_cache: bool
    def __init__(self, *arg, **kw) -> None: ...
    @classmethod
    def and_(cls, *clauses): ...
    @classmethod
    def or_(cls, *clauses): ...
    def self_group(self, against: Any | None = ...): ...

and_: Any
or_: Any

class Tuple(ClauseList, ColumnElement[Any]):
    __visit_name__: str
    type: Any
    def __init__(self, *clauses, **kw) -> None: ...
    def self_group(self, against: Any | None = ...): ...

class Case(ColumnElement[Any]):
    __visit_name__: str
    value: Any
    type: Any
    whens: Any
    else_: Any
    def __init__(self, *whens, **kw) -> None: ...

def literal_column(text, type_: Any | None = ...): ...

class Cast(WrapsColumnExpression, ColumnElement[Any]):
    __visit_name__: str
    type: Any
    clause: Any
    typeclause: Any
    def __init__(self, expression, type_) -> None: ...
    @property
    def wrapped_column_expression(self): ...

class TypeCoerce(WrapsColumnExpression, ColumnElement[Any]):
    __visit_name__: str
    type: Any
    clause: Any
    def __init__(self, expression, type_) -> None: ...
    @HasMemoized.memoized_attribute
    def typed_expression(self): ...
    @property
    def wrapped_column_expression(self): ...
    def self_group(self, against: Any | None = ...): ...

class Extract(ColumnElement[Any]):
    __visit_name__: str
    type: Any
    field: Any
    expr: Any
    def __init__(self, field, expr, **kwargs) -> None: ...

class _label_reference(ColumnElement[Any]):
    __visit_name__: str
    element: Any
    def __init__(self, element) -> None: ...

class _textual_label_reference(ColumnElement[Any]):
    __visit_name__: str
    element: Any
    def __init__(self, element) -> None: ...

class UnaryExpression(ColumnElement[Any]):
    __visit_name__: str
    operator: Any
    modifier: Any
    element: Any
    type: Any
    wraps_column_expression: Any
    def __init__(
        self,
        element,
        operator: Any | None = ...,
        modifier: Any | None = ...,
        type_: Any | None = ...,
        wraps_column_expression: bool = ...,
    ) -> None: ...
    def self_group(self, against: Any | None = ...): ...

class CollectionAggregate(UnaryExpression):
    inherit_cache: bool
    def operate(self, op, *other, **kwargs): ...
    def reverse_operate(self, op, other, **kwargs) -> None: ...

class AsBoolean(WrapsColumnExpression, UnaryExpression):
    inherit_cache: bool
    element: Any
    type: Any
    operator: Any
    negate: Any
    modifier: Any
    wraps_column_expression: bool
    def __init__(self, element, operator, negate) -> None: ...
    @property
    def wrapped_column_expression(self): ...
    def self_group(self, against: Any | None = ...): ...

class BinaryExpression(ColumnElement[Any]):
    __visit_name__: str
    left: Any
    right: Any
    operator: Any
    type: Any
    negate: Any
    modifiers: Any
    def __init__(
        self, left, right, operator, type_: Any | None = ..., negate: Any | None = ..., modifiers: Any | None = ...
    ) -> None: ...
    def __bool__(self): ...
    __nonzero__: Any
    @property
    def is_comparison(self): ...
    def self_group(self, against: Any | None = ...): ...

class Slice(ColumnElement[Any]):
    __visit_name__: str
    start: Any
    stop: Any
    step: Any
    type: Any
    def __init__(self, start, stop, step, _name: Any | None = ...) -> None: ...
    def self_group(self, against: Any | None = ...): ...

class IndexExpression(BinaryExpression):
    inherit_cache: bool

class GroupedElement(ClauseElement):
    __visit_name__: str
    def self_group(self, against: Any | None = ...): ...

class Grouping(GroupedElement, ColumnElement[Any]):
    element: Any
    type: Any
    def __init__(self, element) -> None: ...
    def __getattr__(self, attr): ...

RANGE_UNBOUNDED: Any
RANGE_CURRENT: Any

class Over(ColumnElement[Any]):
    __visit_name__: str
    order_by: Any
    partition_by: Any
    element: Any
    range_: Any
    rows: Any
    def __init__(
        self,
        element,
        partition_by: Any | None = ...,
        order_by: Any | None = ...,
        range_: Any | None = ...,
        rows: Any | None = ...,
    ) -> None: ...
    def __reduce__(self): ...
    @memoized_property
    def type(self): ...

class WithinGroup(ColumnElement[Any]):
    __visit_name__: str
    order_by: Any
    element: Any
    def __init__(self, element, *order_by) -> None: ...
    def __reduce__(self): ...
    def over(
        self, partition_by: Any | None = ..., order_by: Any | None = ..., range_: Any | None = ..., rows: Any | None = ...
    ): ...
    @memoized_property
    def type(self): ...

class FunctionFilter(ColumnElement[Any]):
    __visit_name__: str
    criterion: Any
    func: Any
    def __init__(self, func, *criterion) -> None: ...
    def filter(self, *criterion): ...
    def over(
        self, partition_by: Any | None = ..., order_by: Any | None = ..., range_: Any | None = ..., rows: Any | None = ...
    ): ...
    def self_group(self, against: Any | None = ...): ...
    @memoized_property
    def type(self): ...

class Label(roles.LabeledColumnExprRole, ColumnElement[Any]):
    __visit_name__: str
    name: Any
    key: Any
    def __init__(self, name, element, type_: Any | None = ...) -> None: ...
    def __reduce__(self): ...
    @memoized_property
    def type(self): ...
    @HasMemoized.memoized_attribute
    def element(self): ...
    def self_group(self, against: Any | None = ...): ...
    @property
    def primary_key(self): ...
    @property
    def foreign_keys(self): ...

class NamedColumn(ColumnElement[Any]):
    is_literal: bool
    table: Any
    @memoized_property
    def description(self): ...

class ColumnClause(roles.DDLReferredColumnRole, roles.LabeledColumnExprRole, roles.StrAsPlainColumnRole, Immutable, NamedColumn):
    table: Any
    is_literal: bool
    __visit_name__: str
    onupdate: Any
    default: Any
    server_default: Any
    server_onupdate: Any
    key: Any
    type: Any
    def __init__(self, text, type_: Any | None = ..., is_literal: bool = ..., _selectable: Any | None = ...) -> None: ...
    def get_children(self, column_tables: bool = ..., **kw): ...  # type: ignore[override]
    @property
    def entity_namespace(self): ...

class TableValuedColumn(NamedColumn):
    __visit_name__: str
    scalar_alias: Any
    key: Any
    type: Any
    def __init__(self, scalar_alias, type_) -> None: ...

class CollationClause(ColumnElement[Any]):
    __visit_name__: str
    collation: Any
    def __init__(self, collation) -> None: ...

class _IdentifiedClause(Executable, ClauseElement):
    __visit_name__: str
    ident: Any
    def __init__(self, ident) -> None: ...

class SavepointClause(_IdentifiedClause):
    __visit_name__: str
    inherit_cache: bool

class RollbackToSavepointClause(_IdentifiedClause):
    __visit_name__: str
    inherit_cache: bool

class ReleaseSavepointClause(_IdentifiedClause):
    __visit_name__: str
    inherit_cache: bool

class quoted_name(util.MemoizedSlots, util.text_type):
    quote: Any
    def __new__(cls, value, quote): ...
    def __reduce__(self): ...

class AnnotatedColumnElement(Annotated):
    def __init__(self, element, values) -> None: ...
    @memoized_property
    def name(self): ...
    @memoized_property
    def table(self): ...
    @memoized_property
    def key(self): ...
    @memoized_property
    def info(self): ...

class _truncated_label(quoted_name):
    def __new__(cls, value, quote: Any | None = ...): ...
    def __reduce__(self): ...
    def apply_map(self, map_): ...

class conv(_truncated_label): ...

class _anonymous_label(_truncated_label):
    @classmethod
    def safe_construct(cls, seed, body, enclosing_label: Any | None = ..., sanitize_key: bool = ...): ...
    def __add__(self, other): ...
    def __radd__(self, other): ...
    def apply_map(self, map_): ...



# https://pyinstaller.org/en/stable/advanced-topics.html#the-toc-and-tree-classes
from collections.abc import Iterable, Sequence
from typing import ClassVar
from typing_extensions import Literal, LiteralString, SupportsIndex, TypeAlias

_TypeCode: TypeAlias = Literal["DATA", "BINARY", "EXTENSION", "OPTION"]
_TOCTuple: TypeAlias = tuple[str, str | None, _TypeCode | None]

class TOC(list[_TOCTuple]):
    filenames: set[str]
    def __init__(self, initlist: Iterable[_TOCTuple] | None = ...) -> None: ...
    def append(self, entry: _TOCTuple) -> None: ...
    def insert(self, pos: SupportsIndex, entry: _TOCTuple) -> None: ...
    def extend(self, other: Iterable[_TOCTuple]) -> None: ...

class Target:
    invcnum: ClassVar[int]
    tocfilename: LiteralString
    tocbasename: LiteralString
    dependencies: TOC

class Tree(Target, TOC):
    root: str | None
    prefix: str | None
    excludes: Sequence[str]
    typecode: _TypeCode
    def __init__(
        self, root: str | None = ..., prefix: str | None = ..., excludes: Sequence[str] | None = ..., typecode: _TypeCode = ...
    ) -> None: ...
    def assemble(self) -> None: ...



from html.parser import HTMLParser as HTMLParser
from io import StringIO as StringIO

unicode = str
unichr = chr
basestring = str
interactive_prompt = input

def coerce_string(value): ...
def is_string(value): ...
def is_unicode(value): ...
def on_macos(): ...
def on_windows(): ...



from typing import Any

from ..strategy.asynchronous import AsyncStrategy

class AsyncStreamStrategy(AsyncStrategy):
    can_stream: bool
    line_separator: Any
    all_base64: bool
    stream: Any
    order: Any
    persistent_search_message_id: Any
    streaming: bool
    callback: Any
    events: Any
    def __init__(self, ldap_connection) -> None: ...
    def accumulate_stream(self, message_id, change) -> None: ...
    def get_stream(self): ...
    def set_stream(self, value) -> None: ...



class MenuMargins:
    def __init__(self, top: int = ..., left: int = ..., bottom: int = ..., right: int = ...) -> None: ...
    @property
    def left(self) -> int: ...
    @left.setter
    def left(self, left: int) -> None: ...
    @property
    def right(self) -> int: ...
    @right.setter
    def right(self, right: int) -> None: ...
    @property
    def top(self) -> int: ...
    @top.setter
    def top(self, top: int) -> None: ...
    @property
    def bottom(self) -> int: ...
    @bottom.setter
    def bottom(self, bottom: int) -> None: ...



import logging
from _typeshed import Self
from collections.abc import Collection
from types import TracebackType
from typing import Any, ClassVar, NamedTuple
from typing_extensions import Literal

from .. import exceptions
from ..connectionpool import ConnectionPool
from ..response import HTTPResponse

ConnectTimeoutError = exceptions.ConnectTimeoutError
MaxRetryError = exceptions.MaxRetryError
ProtocolError = exceptions.ProtocolError
ReadTimeoutError = exceptions.ReadTimeoutError
ResponseError = exceptions.ResponseError

log: logging.Logger

class RequestHistory(NamedTuple):
    method: str | None
    url: str | None
    error: Exception | None
    status: int | None
    redirect_location: str | None

class Retry:
    DEFAULT_ALLOWED_METHODS: ClassVar[frozenset[str]]
    RETRY_AFTER_STATUS_CODES: ClassVar[frozenset[int]]
    DEFAULT_REMOVE_HEADERS_ON_REDIRECT: ClassVar[frozenset[str]]
    DEFAULT_BACKOFF_MAX: ClassVar[int]

    total: bool | int | None
    connect: int | None
    read: int | None
    redirect: Literal[True] | int | None
    status: int | None
    other: int | None
    allowed_methods: Collection[str] | Literal[False] | None
    status_forcelist: Collection[int]
    backoff_factor: float
    raise_on_redirect: bool
    raise_on_status: bool
    history: tuple[RequestHistory, ...]
    respect_retry_after_header: bool
    remove_headers_on_redirect: frozenset[str]
    def __init__(
        self,
        total: bool | int | None = ...,
        connect: int | None = ...,
        read: int | None = ...,
        redirect: bool | int | None = ...,
        status: int | None = ...,
        other: int | None = ...,
        allowed_methods: Collection[str] | Literal[False] | None = ...,
        status_forcelist: Collection[int] | None = ...,
        backoff_factor: float = ...,
        raise_on_redirect: bool = ...,
        raise_on_status: bool = ...,
        history: tuple[RequestHistory, ...] | None = ...,
        respect_retry_after_header: bool = ...,
        remove_headers_on_redirect: Collection[str] = ...,
        method_whitelist: Collection[str] | None = ...,
    ) -> None: ...
    def new(self: Self, **kw: Any) -> Self: ...
    @classmethod
    def from_int(
        cls, retries: Retry | bool | int | None, redirect: bool | int | None = ..., default: Retry | bool | int | None = ...
    ) -> Retry: ...
    def get_backoff_time(self) -> float: ...
    def parse_retry_after(self, retry_after: str) -> float: ...
    def get_retry_after(self, response: HTTPResponse) -> float | None: ...
    def sleep_for_retry(self, response: HTTPResponse | None = ...) -> bool: ...
    def sleep(self, response: HTTPResponse | None = ...) -> None: ...
    def is_retry(self, method: str, status_code: int, has_retry_after: bool = ...) -> bool: ...
    def is_exhausted(self) -> bool: ...
    def increment(
        self,
        method: str | None = ...,
        url: str | None = ...,
        response: HTTPResponse | None = ...,
        error: Exception | None = ...,
        _pool: ConnectionPool | None = ...,
        _stacktrace: TracebackType | None = ...,
    ) -> Retry: ...



#!/usr/bin/env python3
"""Test typeshed's third party stubs using stubtest"""

from __future__ import annotations

import argparse
import functools
import os
import subprocess
import sys
import tempfile
import venv
from pathlib import Path
from typing import NoReturn

import tomli
from utils import colored, print_error, print_success_msg


@functools.lru_cache()
def get_mypy_req() -> str:
    with open("requirements-tests.txt", encoding="UTF-8") as f:
        return next(line.strip() for line in f if "mypy" in line)


def run_stubtest(dist: Path, *, verbose: bool = False) -> bool:
    with open(dist / "METADATA.toml", encoding="UTF-8") as f:
        metadata = dict(tomli.loads(f.read()))

    print(f"{dist.name}... ", end="")

    stubtest_meta = metadata.get("tool", {}).get("stubtest", {})
    if stubtest_meta.get("skip", False):
        print(colored("skipping", "yellow"))
        return True

    with tempfile.TemporaryDirectory() as tmp:
        venv_dir = Path(tmp)
        venv.create(venv_dir, with_pip=True, clear=True)

        if sys.platform == "win32":
            pip = venv_dir / "Scripts" / "pip.exe"
            python = venv_dir / "Scripts" / "python.exe"
        else:
            pip = venv_dir / "bin" / "pip"
            python = venv_dir / "bin" / "python"

        pip_exe, python_exe = str(pip), str(python)

        dist_version = metadata["version"]
        extras = stubtest_meta.get("extras", [])
        assert isinstance(dist_version, str)
        assert isinstance(extras, list)
        dist_extras = ", ".join(extras)
        dist_req = f"{dist.name}[{dist_extras}]=={dist_version}"

        # If @tests/requirements-stubtest.txt exists, run "pip install" on it.
        req_path = dist / "@tests" / "requirements-stubtest.txt"
        if req_path.exists():
            try:
                pip_cmd = [pip_exe, "install", "-r", str(req_path)]
                subprocess.run(pip_cmd, check=True, capture_output=True)
            except subprocess.CalledProcessError as e:
                print_command_failure("Failed to install requirements", e)
                return False

        # We need stubtest to be able to import the package, so install mypy into the venv
        # Hopefully mypy continues to not need too many dependencies
        # TODO: Maybe find a way to cache these in CI
        dists_to_install = [dist_req, get_mypy_req()]
        dists_to_install.extend(metadata.get("requires", []))
        pip_cmd = [pip_exe, "install"] + dists_to_install
        try:
            subprocess.run(pip_cmd, check=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            print_command_failure("Failed to install", e)
            return False

        ignore_missing_stub = ["--ignore-missing-stub"] if stubtest_meta.get("ignore_missing_stub", True) else []
        packages_to_check = [d.name for d in dist.iterdir() if d.is_dir() and d.name.isidentifier()]
        modules_to_check = [d.stem for d in dist.iterdir() if d.is_file() and d.suffix == ".pyi"]
        stubtest_cmd = [
            python_exe,
            "-m",
            "mypy.stubtest",
            # Use --custom-typeshed-dir in case we make linked changes to stdlib or _typeshed
            "--custom-typeshed-dir",
            str(dist.parent.parent),
            *ignore_missing_stub,
            *packages_to_check,
            *modules_to_check,
        ]

        # For packages that need a display, we need to pass at least $DISPLAY
        # to stubtest. $DISPLAY is set by xvfb-run in CI.
        #
        # It seems that some other environment variables are needed too,
        # because the CI fails if we pass only os.environ["DISPLAY"]. I didn't
        # "bisect" to see which variables are actually needed.
        stubtest_env = os.environ | {"MYPYPATH": str(dist), "MYPY_FORCE_COLOR": "1"}

        allowlist_path = dist / "@tests/stubtest_allowlist.txt"
        if allowlist_path.exists():
            stubtest_cmd.extend(["--allowlist", str(allowlist_path)])

        try:
            subprocess.run(stubtest_cmd, env=stubtest_env, check=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            print_error("fail")
            print_commands(dist, pip_cmd, stubtest_cmd)
            print_command_output(e)

            print("Ran with the following environment:", file=sys.stderr)
            ret = subprocess.run([pip_exe, "freeze", "--all"], capture_output=True)
            print_command_output(ret)

            if allowlist_path.exists():
                print(
                    f'To fix "unused allowlist" errors, remove the corresponding entries from {allowlist_path}', file=sys.stderr
                )
                print(file=sys.stderr)
            else:
                print(f"Re-running stubtest with --generate-allowlist.\nAdd the following to {allowlist_path}:", file=sys.stderr)
                ret = subprocess.run(stubtest_cmd + ["--generate-allowlist"], env=stubtest_env, capture_output=True)
                print_command_output(ret)

            return False
        else:
            print_success_msg()

    if verbose:
        print_commands(dist, pip_cmd, stubtest_cmd)

    return True


def print_commands(dist: Path, pip_cmd: list[str], stubtest_cmd: list[str]) -> None:
    print(file=sys.stderr)
    print(" ".join(pip_cmd), file=sys.stderr)
    print(f"MYPYPATH={dist}", " ".join(stubtest_cmd), file=sys.stderr)
    print(file=sys.stderr)


def print_command_failure(message: str, e: subprocess.CalledProcessError) -> None:
    print_error("fail")
    print(file=sys.stderr)
    print(message, file=sys.stderr)
    print_command_output(e)


def print_command_output(e: subprocess.CalledProcessError | subprocess.CompletedProcess[bytes]) -> None:
    print(e.stdout.decode(), end="", file=sys.stderr)
    print(e.stderr.decode(), end="", file=sys.stderr)
    print(file=sys.stderr)


def main() -> NoReturn:
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("--num-shards", type=int, default=1)
    parser.add_argument("--shard-index", type=int, default=0)
    parser.add_argument("dists", metavar="DISTRIBUTION", type=str, nargs=argparse.ZERO_OR_MORE)
    args = parser.parse_args()

    typeshed_dir = Path(".").resolve()
    if len(args.dists) == 0:
        dists = sorted((typeshed_dir / "stubs").iterdir())
    else:
        dists = [typeshed_dir / "stubs" / d for d in args.dists]

    result = 0
    for i, dist in enumerate(dists):
        if i % args.num_shards != args.shard_index:
            continue
        if not run_stubtest(dist, verbose=args.verbose):
            result = 1
    sys.exit(result)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



from re import Pattern
from typing import Any

from sqlalchemy import Table
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm import Query

def should_set_tablename(cls: type) -> bool: ...

camelcase_re: Pattern[str]

def camel_to_snake_case(name: str) -> str: ...

class NameMetaMixin(type):
    def __init__(cls, name: str, bases: tuple[type, ...], d: dict[str, Any]) -> None: ...
    def __table_cls__(cls, *args, **kwargs) -> Table | None: ...

class BindMetaMixin(type):
    def __init__(cls, name: str, bases: tuple[type, ...], d: dict[str, Any]) -> None: ...

class DefaultMeta(NameMetaMixin, BindMetaMixin, DeclarativeMeta): ...

class Model:
    query_class: type[Query[Any]] | None
    query: Query[Any] | None



import io
from _typeshed import Incomplete
from typing_extensions import TypeAlias

from .cffi_opcode import *

VERSION_BASE: int
VERSION_EMBEDDED: int
VERSION_CHAR16CHAR32: int
USE_LIMITED_API: Incomplete

class GlobalExpr:
    name: Incomplete
    address: Incomplete
    type_op: Incomplete
    size: Incomplete
    check_value: Incomplete
    def __init__(self, name, address, type_op, size: int = ..., check_value: int = ...) -> None: ...
    def as_c_expr(self): ...
    def as_python_expr(self): ...

class FieldExpr:
    name: Incomplete
    field_offset: Incomplete
    field_size: Incomplete
    fbitsize: Incomplete
    field_type_op: Incomplete
    def __init__(self, name, field_offset, field_size, fbitsize, field_type_op) -> None: ...
    def as_c_expr(self): ...
    def as_python_expr(self) -> None: ...
    def as_field_python_expr(self): ...

class StructUnionExpr:
    name: Incomplete
    type_index: Incomplete
    flags: Incomplete
    size: Incomplete
    alignment: Incomplete
    comment: Incomplete
    first_field_index: Incomplete
    c_fields: Incomplete
    def __init__(self, name, type_index, flags, size, alignment, comment, first_field_index, c_fields) -> None: ...
    def as_c_expr(self): ...
    def as_python_expr(self): ...

class EnumExpr:
    name: Incomplete
    type_index: Incomplete
    size: Incomplete
    signed: Incomplete
    allenums: Incomplete
    def __init__(self, name, type_index, size, signed, allenums) -> None: ...
    def as_c_expr(self): ...
    def as_python_expr(self): ...

class TypenameExpr:
    name: Incomplete
    type_index: Incomplete
    def __init__(self, name, type_index) -> None: ...
    def as_c_expr(self): ...
    def as_python_expr(self): ...

class Recompiler:
    ffi: Incomplete
    module_name: Incomplete
    target_is_python: Incomplete
    def __init__(self, ffi, module_name, target_is_python: bool = ...) -> None: ...
    def needs_version(self, ver) -> None: ...
    cffi_types: Incomplete
    def collect_type_table(self): ...
    ALL_STEPS: Incomplete
    def collect_step_tables(self): ...
    def write_source_to_f(self, f, preamble) -> None: ...
    def write_c_source_to_f(self, f, preamble) -> None: ...
    def write_py_source_to_f(self, f) -> None: ...

NativeIO: TypeAlias = io.StringIO

def make_c_source(ffi, module_name, preamble, target_c_file, verbose: bool = ...): ...
def make_py_source(ffi, module_name, target_py_file, verbose: bool = ...): ...
def recompile(
    ffi,
    module_name,
    preamble,
    tmpdir: str = ...,
    call_c_compiler: bool = ...,
    c_file: Incomplete | None = ...,
    source_extension: str = ...,
    extradir: Incomplete | None = ...,
    compiler_verbose: int = ...,
    target: Incomplete | None = ...,
    debug: Incomplete | None = ...,
    **kwds,
): ...



from typing import Any

from ..utils.search_pattern import wildcard_match as wildcard_match
from .reservoir import Reservoir as Reservoir

class SamplingRule:
    def __init__(
        self,
        name,
        priority,
        rate,
        reservoir_size,
        host: Any | None = ...,
        method: Any | None = ...,
        path: Any | None = ...,
        service: Any | None = ...,
        service_type: Any | None = ...,
    ) -> None: ...
    def match(self, sampling_req): ...
    def is_default(self): ...
    def snapshot_statistics(self): ...
    def merge(self, rule) -> None: ...
    def ever_matched(self): ...
    def time_to_report(self): ...
    def increment_request_count(self) -> None: ...
    def increment_borrow_count(self) -> None: ...
    def increment_sampled_count(self) -> None: ...
    @property
    def rate(self): ...
    @rate.setter
    def rate(self, v) -> None: ...
    @property
    def name(self): ...
    @property
    def priority(self): ...
    @property
    def reservoir(self): ...
    @reservoir.setter
    def reservoir(self, v) -> None: ...
    @property
    def can_borrow(self): ...
    @property
    def request_count(self): ...
    @property
    def borrow_count(self): ...
    @property
    def sampled_count(self): ...



from typing import Any

def prepare_grant_uri(
    uri,
    client_id,
    response_type,
    redirect_uri: Any | None = ...,
    scope: Any | None = ...,
    state: Any | None = ...,
    code_challenge: str | None = ...,
    code_challenge_method: str | None = ...,
    **kwargs,
): ...
def prepare_token_request(
    grant_type, body: str = ..., include_client_id: bool = ..., code_verifier: str | None = ..., **kwargs
): ...
def prepare_token_revocation_request(
    url, token, token_type_hint: str = ..., callback: Any | None = ..., body: str = ..., **kwargs
): ...
def parse_authorization_code_response(uri, state: Any | None = ...): ...
def parse_implicit_response(uri, state: Any | None = ..., scope: Any | None = ...): ...
def parse_token_response(body, scope: Any | None = ...): ...
def validate_token_parameters(params) -> None: ...



from d3dshot.capture_output import CaptureOutputs as CaptureOutputs
from d3dshot.d3dshot import D3DShot as D3DShot

pil_is_available: bool
numpy_is_available: bool
pytorch_is_available: bool
pytorch_gpu_is_available: bool
capture_output_mapping: dict[str, CaptureOutputs]
capture_outputs: list[str]

def determine_available_capture_outputs() -> list[CaptureOutputs]: ...
def create(capture_output: str = ..., frame_buffer_size: int = ...) -> D3DShot: ...



from _typeshed import Incomplete
from collections.abc import Callable
from typing import Any
from typing_extensions import TypeAlias

import docutils.nodes
import docutils.parsers.rst.states
from docutils.languages import _LanguageModule
from docutils.utils import Reporter, SystemMessage

_RoleFn: TypeAlias = Callable[
    [str, str, str, int, docutils.parsers.rst.states.Inliner, dict[str, Any], list[str]],
    tuple[list[docutils.nodes.reference], list[docutils.nodes.reference]],
]

def register_local_role(name: str, role_fn: _RoleFn) -> None: ...
def role(
    role_name: str, language_module: _LanguageModule, lineno: int, reporter: Reporter
) -> tuple[_RoleFn | None, list[SystemMessage]]: ...
def __getattr__(name: str) -> Incomplete: ...



from re import Pattern
from typing import Any, overload

from markdown.core import Markdown

BLOCK_LEVEL_ELEMENTS: Any
STX: str
ETX: str
INLINE_PLACEHOLDER_PREFIX: Any
INLINE_PLACEHOLDER: Any
INLINE_PLACEHOLDER_RE: Pattern[str]
AMP_SUBSTITUTE: Any
HTML_PLACEHOLDER: Any
HTML_PLACEHOLDER_RE: Pattern[str]
TAG_PLACEHOLDER: Any
RTL_BIDI_RANGES: Any

def deprecated(message: str, stacklevel: int = ...): ...
def parseBoolValue(value: object, fail_on_errors: bool = ..., preserve_none: bool = ...) -> bool | None: ...
def code_escape(text: str) -> str: ...
def nearing_recursion_limit() -> bool: ...

class AtomicString(str): ...

class Processor:
    md: Markdown
    def __init__(self, md: Markdown | None = ...) -> None: ...

class HtmlStash:
    html_counter: int = ...
    rawHtmlBlocks: list[str]
    tag_counter: int = ...
    tag_data: list[dict[str, Any]]
    def __init__(self) -> None: ...
    def store(self, html: str) -> str: ...
    def reset(self) -> None: ...
    def get_placeholder(self, key: int) -> str: ...
    def store_tag(self, tag: str, attrs: list[Any], left_index: int, right_index: int) -> str: ...

class Registry:
    def __init__(self) -> None: ...
    def __contains__(self, item: object) -> bool: ...
    def __iter__(self) -> Any: ...
    @overload
    def __getitem__(self, key: slice) -> Registry: ...
    @overload
    def __getitem__(self, key: str | int) -> Any: ...
    def __len__(self) -> int: ...
    def get_index_for_name(self, name: str) -> int: ...
    def register(self, item: Any, name: str, priority: float) -> None: ...
    def deregister(self, name: str, strict: bool = ...) -> None: ...



from collections import OrderedDict
from collections.abc import Callable
from typing import Any, NamedTuple, TypeVar, overload

from psycopg2._ipaddress import register_ipaddress as register_ipaddress
from psycopg2._json import (
    Json as Json,
    register_default_json as register_default_json,
    register_default_jsonb as register_default_jsonb,
    register_json as register_json,
)
from psycopg2._psycopg import (
    REPLICATION_LOGICAL as REPLICATION_LOGICAL,
    REPLICATION_PHYSICAL as REPLICATION_PHYSICAL,
    ReplicationConnection as _replicationConnection,
    ReplicationCursor as _replicationCursor,
    ReplicationMessage as ReplicationMessage,
)
from psycopg2._range import (
    DateRange as DateRange,
    DateTimeRange as DateTimeRange,
    DateTimeTZRange as DateTimeTZRange,
    NumericRange as NumericRange,
    Range as Range,
    RangeAdapter as RangeAdapter,
    RangeCaster as RangeCaster,
    register_range as register_range,
)

from .extensions import connection as _connection, cursor as _cursor, quote_ident as quote_ident

_T_cur = TypeVar("_T_cur", bound=_cursor)

class DictCursorBase(_cursor):
    def __init__(self, *args, **kwargs) -> None: ...

class DictConnection(_connection):
    @overload
    def cursor(self, name: str | bytes | None = ..., *, withhold: bool = ..., scrollable: bool | None = ...) -> DictCursor: ...
    @overload
    def cursor(
        self,
        name: str | bytes | None = ...,
        *,
        cursor_factory: Callable[..., _T_cur],
        withhold: bool = ...,
        scrollable: bool | None = ...,
    ) -> _T_cur: ...
    @overload
    def cursor(
        self, name: str | bytes | None, cursor_factory: Callable[..., _T_cur], withhold: bool = ..., scrollable: bool | None = ...
    ) -> _T_cur: ...

class DictCursor(DictCursorBase):
    def __init__(self, *args, **kwargs) -> None: ...
    index: Any
    def execute(self, query, vars: Any | None = ...): ...
    def callproc(self, procname, vars: Any | None = ...): ...
    def fetchone(self) -> DictRow | None: ...  # type: ignore[override]
    def fetchmany(self, size: int | None = ...) -> list[DictRow]: ...  # type: ignore[override]
    def fetchall(self) -> list[DictRow]: ...  # type: ignore[override]
    def __next__(self) -> DictRow: ...  # type: ignore[override]

class DictRow(list[Any]):
    def __init__(self, cursor) -> None: ...
    def __getitem__(self, x): ...
    def __setitem__(self, x, v) -> None: ...
    def items(self): ...
    def keys(self): ...
    def values(self): ...
    def get(self, x, default: Any | None = ...): ...
    def copy(self): ...
    def __contains__(self, x): ...
    def __reduce__(self): ...

class RealDictConnection(_connection):
    @overload
    def cursor(
        self, name: str | bytes | None = ..., *, withhold: bool = ..., scrollable: bool | None = ...
    ) -> RealDictCursor: ...
    @overload
    def cursor(
        self,
        name: str | bytes | None = ...,
        *,
        cursor_factory: Callable[..., _T_cur],
        withhold: bool = ...,
        scrollable: bool | None = ...,
    ) -> _T_cur: ...
    @overload
    def cursor(
        self, name: str | bytes | None, cursor_factory: Callable[..., _T_cur], withhold: bool = ..., scrollable: bool | None = ...
    ) -> _T_cur: ...

class RealDictCursor(DictCursorBase):
    def __init__(self, *args, **kwargs) -> None: ...
    column_mapping: Any
    def execute(self, query, vars: Any | None = ...): ...
    def callproc(self, procname, vars: Any | None = ...): ...
    def fetchone(self) -> RealDictRow | None: ...  # type: ignore[override]
    def fetchmany(self, size: int | None = ...) -> list[RealDictRow]: ...  # type: ignore[override]
    def fetchall(self) -> list[RealDictRow]: ...  # type: ignore[override]
    def __next__(self) -> RealDictRow: ...  # type: ignore[override]

class RealDictRow(OrderedDict[Any, Any]):
    def __init__(self, *args, **kwargs) -> None: ...
    def __setitem__(self, key, value) -> None: ...

class NamedTupleConnection(_connection):
    @overload
    def cursor(
        self, name: str | bytes | None = ..., *, withhold: bool = ..., scrollable: bool | None = ...
    ) -> NamedTupleCursor: ...
    @overload
    def cursor(
        self,
        name: str | bytes | None = ...,
        *,
        cursor_factory: Callable[..., _T_cur],
        withhold: bool = ...,
        scrollable: bool | None = ...,
    ) -> _T_cur: ...
    @overload
    def cursor(
        self, name: str | bytes | None, cursor_factory: Callable[..., _T_cur], withhold: bool = ..., scrollable: bool | None = ...
    ) -> _T_cur: ...

class NamedTupleCursor(_cursor):
    Record: Any
    MAX_CACHE: int
    def execute(self, query, vars: Any | None = ...): ...
    def executemany(self, query, vars): ...
    def callproc(self, procname, vars: Any | None = ...): ...
    def fetchone(self) -> NamedTuple | None: ...  # type: ignore[override]
    def fetchmany(self, size: int | None = ...) -> list[NamedTuple]: ...  # type: ignore[override]
    def fetchall(self) -> list[NamedTuple]: ...  # type: ignore[override]
    def __next__(self) -> NamedTuple: ...  # type: ignore[override]

class LoggingConnection(_connection):
    log: Any
    def initialize(self, logobj) -> None: ...
    def filter(self, msg, curs): ...
    def cursor(self, *args, **kwargs): ...

class LoggingCursor(_cursor):
    def execute(self, query, vars: Any | None = ...): ...
    def callproc(self, procname, vars: Any | None = ...): ...

class MinTimeLoggingConnection(LoggingConnection):
    def initialize(self, logobj, mintime: int = ...) -> None: ...
    def filter(self, msg, curs): ...
    def cursor(self, *args, **kwargs): ...

class MinTimeLoggingCursor(LoggingCursor):
    timestamp: Any
    def execute(self, query, vars: Any | None = ...): ...
    def callproc(self, procname, vars: Any | None = ...): ...

class LogicalReplicationConnection(_replicationConnection):
    def __init__(self, *args, **kwargs) -> None: ...

class PhysicalReplicationConnection(_replicationConnection):
    def __init__(self, *args, **kwargs) -> None: ...

class StopReplication(Exception): ...

class ReplicationCursor(_replicationCursor):
    def create_replication_slot(self, slot_name, slot_type: Any | None = ..., output_plugin: Any | None = ...) -> None: ...
    def drop_replication_slot(self, slot_name) -> None: ...
    def start_replication(
        self,
        slot_name: Any | None = ...,
        slot_type: Any | None = ...,
        start_lsn: int = ...,
        timeline: int = ...,
        options: Any | None = ...,
        decode: bool = ...,
        status_interval: int = ...,
    ) -> None: ...
    def fileno(self): ...

class UUID_adapter:
    def __init__(self, uuid) -> None: ...
    def __conform__(self, proto): ...
    def getquoted(self): ...

def register_uuid(oids: Any | None = ..., conn_or_curs: Any | None = ...): ...

class Inet:
    addr: Any
    def __init__(self, addr) -> None: ...
    def prepare(self, conn) -> None: ...
    def getquoted(self): ...
    def __conform__(self, proto): ...

def register_inet(oid: Any | None = ..., conn_or_curs: Any | None = ...): ...
def wait_select(conn) -> None: ...

class HstoreAdapter:
    wrapped: Any
    def __init__(self, wrapped) -> None: ...
    conn: Any
    getquoted: Any
    def prepare(self, conn) -> None: ...
    @classmethod
    def parse(cls, s, cur, _bsdec=...): ...
    @classmethod
    def parse_unicode(cls, s, cur): ...
    @classmethod
    def get_oids(cls, conn_or_curs): ...

def register_hstore(
    conn_or_curs, globally: bool = ..., unicode: bool = ..., oid: Any | None = ..., array_oid: Any | None = ...
) -> None: ...

class CompositeCaster:
    name: Any
    schema: Any
    oid: Any
    array_oid: Any
    attnames: Any
    atttypes: Any
    typecaster: Any
    array_typecaster: Any
    def __init__(self, name, oid, attrs, array_oid: Any | None = ..., schema: Any | None = ...) -> None: ...
    def parse(self, s, curs): ...
    def make(self, values): ...
    @classmethod
    def tokenize(cls, s): ...

def register_composite(name, conn_or_curs, globally: bool = ..., factory: Any | None = ...): ...
def execute_batch(cur, sql, argslist, page_size: int = ...) -> None: ...
def execute_values(cur, sql, argslist, template: Any | None = ..., page_size: int = ..., fetch: bool = ...): ...



import datetime
from re import Pattern
from types import TracebackType
from typing import Any, NamedTuple

class SizeUnit(NamedTuple):
    divider: int
    symbol: str
    name: str

class CombinedUnit(NamedTuple):
    decimal: SizeUnit
    binary: SizeUnit

disk_size_units: Any
length_size_units: Any
time_units: Any

def coerce_boolean(value: object) -> bool: ...
def coerce_pattern(value: str | Pattern[str], flags: int = ...) -> Pattern[str]: ...
def coerce_seconds(value: float | datetime.timedelta) -> float: ...
def format_size(num_bytes: float, keep_width: bool = ..., binary: bool = ...) -> str: ...
def parse_size(size: str, binary: bool = ...) -> int: ...
def format_length(num_metres: float, keep_width: bool = ...) -> str: ...
def parse_length(length: str) -> float: ...
def format_number(number: float, num_decimals: int = ...) -> str: ...
def round_number(count: float, keep_width: bool = ...) -> str: ...
def format_timespan(num_seconds: float | datetime.timedelta, detailed: bool = ..., max_units: int = ...) -> str: ...
def parse_timespan(timespan: str) -> float: ...
def parse_date(datestring: str) -> tuple[int, int, int, int, int, int]: ...
def format_path(pathname: str) -> str: ...
def parse_path(pathname: str) -> str: ...

class Timer:
    monotonic: bool
    resumable: bool
    start_time: float
    total_time: float
    def __init__(self, start_time: Any | None = ..., resumable: bool = ...) -> None: ...
    def __enter__(self): ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None = ...,
        exc_value: BaseException | None = ...,
        traceback: TracebackType | None = ...,
    ) -> None: ...
    def sleep(self, seconds: float) -> None: ...
    @property
    def elapsed_time(self): ...
    @property
    def rounded(self): ...

class InvalidDate(Exception): ...
class InvalidSize(Exception): ...
class InvalidLength(Exception): ...
class InvalidTimespan(Exception): ...



from .base import NonRecursiveTreeWalker

class TreeWalker(NonRecursiveTreeWalker):
    def getNodeDetails(self, node): ...
    def getFirstChild(self, node): ...
    def getNextSibling(self, node): ...
    def getParentNode(self, node): ...



from _typeshed import Incomplete

class VGenericEngine:
    verifier: Incomplete
    ffi: Incomplete
    export_symbols: Incomplete
    def __init__(self, verifier) -> None: ...
    def patch_extension_kwds(self, kwds) -> None: ...
    def find_module(self, module_name, path, so_suffixes): ...
    def collect_types(self) -> None: ...
    def write_source_to_f(self) -> None: ...
    def load_library(self, flags: int = ...): ...

cffimod_header: str



from collections.abc import Callable
from threading import Event
from typing_extensions import TypeAlias

from paramiko.pkey import PKey
from paramiko.ssh_gss import _SSH_GSSAuth
from paramiko.transport import Transport

_InteractiveCallback: TypeAlias = Callable[[str, str, list[tuple[str, bool]]], list[str]]

class AuthHandler:
    transport: Transport
    username: str | None
    authenticated: bool
    auth_event: Event | None
    auth_method: str
    banner: str | None
    password: str | None
    private_key: PKey | None
    interactive_handler: _InteractiveCallback | None
    submethods: str | None
    auth_username: str | None
    auth_fail_count: int
    gss_host: str | None
    gss_deleg_creds: bool
    def __init__(self, transport: Transport) -> None: ...
    def is_authenticated(self) -> bool: ...
    def get_username(self) -> str | None: ...
    def auth_none(self, username: str, event: Event) -> None: ...
    def auth_publickey(self, username: str, key: PKey, event: Event) -> None: ...
    def auth_password(self, username: str, password: str, event: Event) -> None: ...
    def auth_interactive(self, username: str, handler: _InteractiveCallback, event: Event, submethods: str = ...) -> None: ...
    def auth_gssapi_with_mic(self, username: str, gss_host: str, gss_deleg_creds: bool, event: Event) -> None: ...
    def auth_gssapi_keyex(self, username: str, event: Event) -> None: ...
    def abort(self) -> None: ...
    def wait_for_response(self, event: Event) -> list[str]: ...

class GssapiWithMicAuthHandler:
    method: str
    sshgss: _SSH_GSSAuth
    def __init__(self, delegate: AuthHandler, sshgss: _SSH_GSSAuth) -> None: ...
    def abort(self) -> None: ...
    @property
    def transport(self) -> Transport: ...
    @property
    def auth_username(self) -> str: ...
    @property
    def gss_host(self) -> str: ...



from _typeshed import Self
from re import Pattern

class Version:
    def __init__(self, vstring: str | None = ...) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __lt__(self: Self, other: Self | str) -> bool: ...
    def __le__(self: Self, other: Self | str) -> bool: ...
    def __gt__(self: Self, other: Self | str) -> bool: ...
    def __ge__(self: Self, other: Self | str) -> bool: ...

class StrictVersion(Version):
    version_re: Pattern[str]
    version: tuple[int, int, int]
    prerelease: tuple[str, int] | None
    def __init__(self, vstring: str | None = ...) -> None: ...
    def parse(self: Self, vstring: str) -> Self: ...
    def _cmp(self: Self, other: Self | str) -> bool: ...

class LooseVersion(Version):
    component_re: Pattern[str]
    vstring: str
    version: tuple[str | int, ...]
    def __init__(self, vstring: str | None = ...) -> None: ...
    def parse(self: Self, vstring: str) -> Self: ...
    def _cmp(self: Self, other: Self | str) -> bool: ...



from typing import Any

from .base import Client as Client

class LegacyApplicationClient(Client):
    grant_type: str
    def __init__(self, client_id, **kwargs) -> None: ...
    def prepare_request_body(  # type: ignore[override]
        self, username, password, body: str = ..., scope: Any | None = ..., include_client_id: bool = ..., **kwargs
    ): ...



from typing import Any

from braintree.amex_express_checkout_card import AmexExpressCheckoutCard as AmexExpressCheckoutCard
from braintree.android_pay_card import AndroidPayCard as AndroidPayCard
from braintree.apple_pay_card import ApplePayCard as ApplePayCard
from braintree.credit_card import CreditCard as CreditCard
from braintree.error_result import ErrorResult as ErrorResult
from braintree.europe_bank_account import EuropeBankAccount as EuropeBankAccount
from braintree.exceptions.not_found_error import NotFoundError as NotFoundError
from braintree.ids_search import IdsSearch as IdsSearch
from braintree.masterpass_card import MasterpassCard as MasterpassCard
from braintree.payment_method import PaymentMethod as PaymentMethod
from braintree.payment_method_nonce import PaymentMethodNonce as PaymentMethodNonce
from braintree.payment_method_parser import parse_payment_method as parse_payment_method
from braintree.paypal_account import PayPalAccount as PayPalAccount
from braintree.resource import Resource as Resource
from braintree.resource_collection import ResourceCollection as ResourceCollection
from braintree.samsung_pay_card import SamsungPayCard as SamsungPayCard
from braintree.successful_result import SuccessfulResult as SuccessfulResult
from braintree.unknown_payment_method import UnknownPaymentMethod as UnknownPaymentMethod
from braintree.us_bank_account import UsBankAccount as UsBankAccount
from braintree.venmo_account import VenmoAccount as VenmoAccount
from braintree.visa_checkout_card import VisaCheckoutCard as VisaCheckoutCard

class PaymentMethodGateway:
    gateway: Any
    config: Any
    def __init__(self, gateway) -> None: ...
    def create(self, params: Any | None = ...): ...
    def find(self, payment_method_token): ...
    def update(self, payment_method_token, params): ...
    def delete(self, payment_method_token, options: Any | None = ...): ...
    options: Any
    def grant(self, payment_method_token, options: Any | None = ...): ...
    def revoke(self, payment_method_token): ...



from _typeshed import ReadableBuffer, SupportsRead
from collections.abc import Callable
from pyexpat import errors as errors, model as model
from typing import Any
from typing_extensions import TypeAlias, final

EXPAT_VERSION: str  # undocumented
version_info: tuple[int, int, int]  # undocumented
native_encoding: str  # undocumented
features: list[tuple[str, int]]  # undocumented

class ExpatError(Exception):
    code: int
    lineno: int
    offset: int

error = ExpatError

XML_PARAM_ENTITY_PARSING_NEVER: int
XML_PARAM_ENTITY_PARSING_UNLESS_STANDALONE: int
XML_PARAM_ENTITY_PARSING_ALWAYS: int

_Model: TypeAlias = tuple[int, int, str | None, tuple[Any, ...]]

@final
class XMLParserType:
    def Parse(self, __data: str | ReadableBuffer, __isfinal: bool = ...) -> int: ...
    def ParseFile(self, __file: SupportsRead[bytes]) -> int: ...
    def SetBase(self, __base: str) -> None: ...
    def GetBase(self) -> str | None: ...
    def GetInputContext(self) -> bytes | None: ...
    def ExternalEntityParserCreate(self, __context: str | None, __encoding: str = ...) -> XMLParserType: ...
    def SetParamEntityParsing(self, __flag: int) -> int: ...
    def UseForeignDTD(self, __flag: bool = ...) -> None: ...
    @property
    def intern(self) -> dict[str, str]: ...
    buffer_size: int
    buffer_text: bool
    buffer_used: int
    namespace_prefixes: bool  # undocumented
    ordered_attributes: bool
    specified_attributes: bool
    ErrorByteIndex: int
    ErrorCode: int
    ErrorColumnNumber: int
    ErrorLineNumber: int
    CurrentByteIndex: int
    CurrentColumnNumber: int
    CurrentLineNumber: int
    XmlDeclHandler: Callable[[str, str | None, int], Any] | None
    StartDoctypeDeclHandler: Callable[[str, str | None, str | None, bool], Any] | None
    EndDoctypeDeclHandler: Callable[[], Any] | None
    ElementDeclHandler: Callable[[str, _Model], Any] | None
    AttlistDeclHandler: Callable[[str, str, str, str | None, bool], Any] | None
    StartElementHandler: Callable[[str, dict[str, str]], Any] | Callable[[str, list[str]], Any] | Callable[
        [str, dict[str, str], list[str]], Any
    ] | None
    EndElementHandler: Callable[[str], Any] | None
    ProcessingInstructionHandler: Callable[[str, str], Any] | None
    CharacterDataHandler: Callable[[str], Any] | None
    UnparsedEntityDeclHandler: Callable[[str, str | None, str, str | None, str], Any] | None
    EntityDeclHandler: Callable[[str, bool, str | None, str | None, str, str | None, str | None], Any] | None
    NotationDeclHandler: Callable[[str, str | None, str, str | None], Any] | None
    StartNamespaceDeclHandler: Callable[[str, str], Any] | None
    EndNamespaceDeclHandler: Callable[[str], Any] | None
    CommentHandler: Callable[[str], Any] | None
    StartCdataSectionHandler: Callable[[], Any] | None
    EndCdataSectionHandler: Callable[[], Any] | None
    DefaultHandler: Callable[[str], Any] | None
    DefaultHandlerExpand: Callable[[str], Any] | None
    NotStandaloneHandler: Callable[[], int] | None
    ExternalEntityRefHandler: Callable[[str, str | None, str | None, str | None], int] | None
    SkippedEntityHandler: Callable[[str, bool], Any] | None

def ErrorString(__code: int) -> str: ...

# intern is undocumented
def ParserCreate(
    encoding: str | None = ..., namespace_separator: str | None = ..., intern: dict[str, Any] | None = ...
) -> XMLParserType: ...



from typing import Any, List, Union
from typing_extensions import Literal, assert_type


class Foo:
    def __add__(self, other: Any) -> "Foo":
        return Foo()


class Bar:
    def __radd__(self, other: Any) -> "Bar":
        return Bar()


class Baz:
    def __add__(self, other: Any) -> "Baz":
        return Baz()

    def __radd__(self, other: Any) -> "Baz":
        return Baz()


literal_list: List[Literal[0, 1]] = [0, 1, 1]

assert_type(sum([2, 4]), int)
assert_type(sum([3, 5], 4), int)

assert_type(sum([True, False]), int)
assert_type(sum([True, False], True), int)
assert_type(sum(literal_list), int)

assert_type(sum([["foo"], ["bar"]], ["baz"]), List[str])

assert_type(sum([Foo(), Foo()], Foo()), Foo)
assert_type(sum([Baz(), Baz()]), Union[Baz, Literal[0]])

# mypy and pyright infer the types differently for these, so we can't use assert_type
# Just test that no error is emitted for any of these
sum([("foo",), ("bar", "baz")], ())  # mypy: `tuple[str, ...]`; pyright: `tuple[()] | tuple[str] | tuple[str, str]`
sum([5.6, 3.2])  # mypy: `float`; pyright: `float | Literal[0]`
sum([2.5, 5.8], 5)  # mypy: `float`; pyright: `float | int`

# These all fail at runtime
sum("abcde")  # type: ignore
sum([["foo"], ["bar"]])  # type: ignore
sum([("foo",), ("bar", "baz")])  # type: ignore
sum([Foo(), Foo()])  # type: ignore
sum([Bar(), Bar()], Bar())  # type: ignore
sum([Bar(), Bar()])  # type: ignore

# TODO: these pass pyright with the current stubs, but mypy erroneously emits an error:
# sum([3, Fraction(7, 22), complex(8, 0), 9.83])
# sum([3, Decimal('0.98')])



from typing import List, Union
from typing_extensions import assert_type


# list.__add__ example from #8292
class Foo:
    def asd(self) -> int:
        return 1


class Bar:
    def asd(self) -> int:
        return 2


combined = [Foo()] + [Bar()]
assert_type(combined, List[Union[Foo, Bar]])
for item in combined:
    assert_type(item.asd(), int)



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class WorkbookProperties(Serialisable):
    tagname: str
    date1904: Any
    dateCompatibility: Any
    showObjects: Any
    showBorderUnselectedTables: Any
    filterPrivacy: Any
    promptedSolutions: Any
    showInkAnnotation: Any
    backupFile: Any
    saveExternalLinkValues: Any
    updateLinks: Any
    codeName: Any
    hidePivotFieldList: Any
    showPivotChartFilter: Any
    allowRefreshQuery: Any
    publishItems: Any
    checkCompatibility: Any
    autoCompressPictures: Any
    refreshAllConnections: Any
    defaultThemeVersion: Any
    def __init__(
        self,
        date1904: Any | None = ...,
        dateCompatibility: Any | None = ...,
        showObjects: Any | None = ...,
        showBorderUnselectedTables: Any | None = ...,
        filterPrivacy: Any | None = ...,
        promptedSolutions: Any | None = ...,
        showInkAnnotation: Any | None = ...,
        backupFile: Any | None = ...,
        saveExternalLinkValues: Any | None = ...,
        updateLinks: Any | None = ...,
        codeName: Any | None = ...,
        hidePivotFieldList: Any | None = ...,
        showPivotChartFilter: Any | None = ...,
        allowRefreshQuery: Any | None = ...,
        publishItems: Any | None = ...,
        checkCompatibility: Any | None = ...,
        autoCompressPictures: Any | None = ...,
        refreshAllConnections: Any | None = ...,
        defaultThemeVersion: Any | None = ...,
    ) -> None: ...

class CalcProperties(Serialisable):
    tagname: str
    calcId: Any
    calcMode: Any
    fullCalcOnLoad: Any
    refMode: Any
    iterate: Any
    iterateCount: Any
    iterateDelta: Any
    fullPrecision: Any
    calcCompleted: Any
    calcOnSave: Any
    concurrentCalc: Any
    concurrentManualCount: Any
    forceFullCalc: Any
    def __init__(
        self,
        calcId: int = ...,
        calcMode: Any | None = ...,
        fullCalcOnLoad: bool = ...,
        refMode: Any | None = ...,
        iterate: Any | None = ...,
        iterateCount: Any | None = ...,
        iterateDelta: Any | None = ...,
        fullPrecision: Any | None = ...,
        calcCompleted: Any | None = ...,
        calcOnSave: Any | None = ...,
        concurrentCalc: Any | None = ...,
        concurrentManualCount: Any | None = ...,
        forceFullCalc: Any | None = ...,
    ) -> None: ...

class FileVersion(Serialisable):
    tagname: str
    appName: Any
    lastEdited: Any
    lowestEdited: Any
    rupBuild: Any
    codeName: Any
    def __init__(
        self,
        appName: Any | None = ...,
        lastEdited: Any | None = ...,
        lowestEdited: Any | None = ...,
        rupBuild: Any | None = ...,
        codeName: Any | None = ...,
    ) -> None: ...



import contextlib
from collections.abc import Callable, Iterable, Sequence
from datetime import datetime
from typing import NamedTuple, SupportsInt, TypeVar
from typing_extensions import ParamSpec, SupportsIndex, TypeAlias

from pyscreeze import (
    locate as locate,
    locateAll as locateAll,
    locateAllOnScreen as locateAllOnScreen,
    locateCenterOnScreen as locateCenterOnScreen,
    locateOnScreen as locateOnScreen,
    locateOnWindow as locateOnWindow,
)

class PyAutoGUIException(Exception): ...
class FailSafeException(PyAutoGUIException): ...
class ImageNotFoundException(PyAutoGUIException): ...

_P = ParamSpec("_P")
_R = TypeVar("_R")
_NormalizeableXArg: TypeAlias = str | SupportsInt | Sequence[SupportsInt]

def raisePyAutoGUIImageNotFoundException(wrappedFunction: Callable[_P, _R]) -> Callable[_P, _R]: ...
def mouseInfo() -> None: ...
def useImageNotFoundException(value: bool | None = ...) -> None: ...

KEY_NAMES: list[str]
KEYBOARD_KEYS: list[str]
LEFT: str
MIDDLE: str
RIGHT: str
PRIMARY: str
SECONDARY: str
QWERTY: str
QWERTZ: str

def isShiftCharacter(character: str) -> bool: ...

MINIMUM_DURATION: float
MINIMUM_SLEEP: float
PAUSE: float
DARWIN_CATCH_UP_TIME: float
FAILSAFE: bool
FAILSAFE_POINTS: list[tuple[int, int]]
LOG_SCREENSHOTS: bool
LOG_SCREENSHOTS_LIMIT: int
G_LOG_SCREENSHOTS_FILENAMES: list[str]

class Point(NamedTuple):
    x: float
    y: float

class Size(NamedTuple):
    width: int
    height: int

def getPointOnLine(x1: float, y1: float, x2: float, y2: float, n: float) -> tuple[float, float]: ...
def linear(n: float) -> float: ...
def position(x: int | None = ..., y: int | None = ...) -> Point: ...
def size() -> Size: ...
def onScreen(x: _NormalizeableXArg | None, y: SupportsInt | None = ...) -> bool: ...
def mouseDown(
    x: _NormalizeableXArg | None = ...,
    y: SupportsInt | None = ...,
    # Docstring says `button` can also be `int`, but `.lower()` is called unconditionally in `_normalizeButton()`
    button: str = ...,
    duration: float = ...,
    tween: Callable[[float], float] = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
) -> None: ...
def mouseUp(
    x: _NormalizeableXArg | None = ...,
    y: SupportsInt | None = ...,
    # Docstring says `button` can also be `int`, but `.lower()` is called unconditionally in `_normalizeButton()`
    button: str = ...,
    duration: float = ...,
    tween: Callable[[float], float] = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
) -> None: ...
def click(
    x: _NormalizeableXArg | None = ...,
    y: SupportsInt | None = ...,
    clicks: SupportsIndex = ...,
    interval: float = ...,
    # Docstring says `button` can also be `int`, but `.lower()` is called unconditionally in `_normalizeButton()`
    button: str = ...,
    duration: float = ...,
    tween: Callable[[float], float] = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
) -> None: ...
def leftClick(
    x: _NormalizeableXArg | None = ...,
    y: SupportsInt | None = ...,
    interval: float = ...,
    duration: float = ...,
    tween: Callable[[float], float] = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
) -> None: ...
def rightClick(
    x: _NormalizeableXArg | None = ...,
    y: SupportsInt | None = ...,
    interval: float = ...,
    duration: float = ...,
    tween: Callable[[float], float] = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
) -> None: ...
def middleClick(
    x: _NormalizeableXArg | None = ...,
    y: SupportsInt | None = ...,
    interval: float = ...,
    duration: float = ...,
    tween: Callable[[float], float] = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
) -> None: ...
def doubleClick(
    x: _NormalizeableXArg | None = ...,
    y: SupportsInt | None = ...,
    interval: float = ...,
    # Docstring says `button` can also be `int`, but `.lower()` is called unconditionally in `_normalizeButton()`
    button: str = ...,
    duration: float = ...,
    tween: Callable[[float], float] = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
) -> None: ...
def tripleClick(
    x: _NormalizeableXArg | None = ...,
    y: SupportsInt | None = ...,
    interval: float = ...,
    # Docstring says `button` can also be `int`, but `.lower()` is called unconditionally in `_normalizeButton()`
    button: str = ...,
    duration: float = ...,
    tween: Callable[[float], float] = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
) -> None: ...
def scroll(
    clicks: float,
    x: _NormalizeableXArg | None = ...,
    y: SupportsInt | None = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
) -> None: ...
def hscroll(
    clicks: float,
    x: _NormalizeableXArg | None = ...,
    y: SupportsInt | None = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
) -> None: ...
def vscroll(
    clicks: float,
    x: _NormalizeableXArg | None = ...,
    y: SupportsInt | None = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
) -> None: ...
def moveTo(
    x: _NormalizeableXArg | None = ...,
    y: SupportsInt | None = ...,
    duration: float = ...,
    tween: Callable[[float], float] = ...,
    logScreenshot: bool = ...,
    _pause: bool = ...,
) -> None: ...
def moveRel(
    xOffset: _NormalizeableXArg | None = ...,
    yOffset: SupportsInt | None = ...,
    duration: float = ...,
    tween: Callable[[float], float] = ...,
    logScreenshot: bool = ...,
    _pause: bool = ...,
) -> None: ...

move = moveRel

def dragTo(
    x: _NormalizeableXArg | None = ...,
    y: SupportsInt | None = ...,
    duration: float = ...,
    tween: Callable[[float], float] = ...,
    # Docstring says `button` can also be `int`, but `.lower()` is called unconditionally in `_normalizeButton()`
    button: str = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
    mouseDownUp: bool = ...,
) -> None: ...
def dragRel(
    xOffset: _NormalizeableXArg | None = ...,
    yOffset: SupportsInt | None = ...,
    duration: float = ...,
    tween: Callable[[float], float] = ...,
    # Docstring says `button` can also be `int`, but `.lower()` is called unconditionally in `_normalizeButton()`
    button: str = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
    mouseDownUp: bool = ...,
) -> None: ...

drag = dragRel

def isValidKey(key: str) -> bool: ...
def keyDown(key: str, logScreenshot: bool | None = ..., _pause: bool = ...) -> None: ...
def keyUp(key: str, logScreenshot: bool | None = ..., _pause: bool = ...) -> None: ...
def press(
    keys: str | Iterable[str],
    presses: SupportsIndex = ...,
    interval: float = ...,
    logScreenshot: bool | None = ...,
    _pause: bool = ...,
) -> None: ...
def hold(
    keys: str | Iterable[str], logScreenshot: bool | None = ..., _pause: bool = ...
) -> contextlib._GeneratorContextManager[None]: ...
def typewrite(
    message: str | Sequence[str], interval: float = ..., logScreenshot: bool | None = ..., _pause: bool = ...
) -> None: ...

write = typewrite

def hotkey(*args: str, logScreenshot: bool | None = ..., interval: float = ...) -> None: ...
def failSafeCheck() -> None: ...
def displayMousePosition(xOffset: float = ..., yOffset: float = ...) -> None: ...
def sleep(seconds: float) -> None: ...
def countdown(seconds: SupportsIndex) -> None: ...
def run(commandStr: str, _ssCount: Sequence[int] | None = ...) -> None: ...
def printInfo(dontPrint: bool = ...) -> str: ...
def getInfo() -> tuple[str, str, str, str, Size, datetime]: ...



from typing import Any

def extended_operation(request_name, request_value: Any | None = ..., no_encode: Any | None = ...): ...
def extended_request_to_dict(request): ...
def extended_response_to_dict(response): ...
def intermediate_response_to_dict(response): ...
def extended_response_to_dict_fast(response): ...
def intermediate_response_to_dict_fast(response): ...



from typing import Any

class ObjectDef:
    def __init__(
        self,
        object_class: Any | None = ...,
        schema: Any | None = ...,
        custom_validator: Any | None = ...,
        auxiliary_class: Any | None = ...,
    ) -> None: ...
    def __getitem__(self, item): ...
    def __getattr__(self, item): ...
    def __setattr__(self, key, value) -> None: ...
    def __iadd__(self, other): ...
    def __isub__(self, other): ...
    def __iter__(self): ...
    def __len__(self): ...
    def __bool__(self): ...
    def __contains__(self, item): ...
    def add_from_schema(self, attribute_name, mandatory: bool = ...) -> None: ...
    def add_attribute(self, definition: Any | None = ...) -> None: ...
    def remove_attribute(self, item) -> None: ...
    def clear_attributes(self) -> None: ...



from _typeshed import Self
from abc import ABCMeta, abstractmethod
from typing import Any, ClassVar
from typing_extensions import Literal

class PasswordHash(metaclass=ABCMeta):
    is_disabled: ClassVar[bool]
    truncate_size: ClassVar[int | None]
    truncate_error: ClassVar[bool]
    truncate_verify_reject: ClassVar[bool]
    @classmethod
    @abstractmethod
    def hash(cls, secret: str | bytes, **setting_and_context_kwds) -> str: ...
    @classmethod
    def encrypt(cls, secret: str | bytes, **kwds) -> str: ...
    @classmethod
    @abstractmethod
    def verify(cls, secret: str | bytes, hash: str | bytes, **context_kwds): ...
    @classmethod
    @abstractmethod
    def using(cls: type[Self], relaxed: bool = ..., **kwds: Any) -> type[Self]: ...
    @classmethod
    def needs_update(cls, hash: str, secret: str | bytes | None = ...) -> bool: ...
    @classmethod
    @abstractmethod
    def identify(cls, hash: str | bytes) -> bool: ...
    @classmethod
    def genconfig(cls, **setting_kwds: Any) -> str: ...
    @classmethod
    def genhash(cls, secret: str | bytes, config: str, **context: Any) -> str: ...
    deprecated: bool

class DisabledHash(PasswordHash, metaclass=ABCMeta):
    is_disabled: ClassVar[Literal[True]]
    @classmethod
    def disable(cls, hash: str | None = ...) -> str: ...
    @classmethod
    def enable(cls, hash: str) -> str: ...



from collections.abc import Callable, Iterable
from typing import TypeVar

from redis.backoff import AbstractBackoff

_T = TypeVar("_T")

class Retry:
    def __init__(self, backoff: AbstractBackoff, retries: int, supported_errors: tuple[type[Exception], ...] = ...) -> None: ...
    def update_supported_errors(self, specified_errors: Iterable[type[Exception]]) -> None: ...
    def call_with_retry(self, do: Callable[[], _T], fail: Callable[[Exception], object]) -> _T: ...



from typing import Any

from braintree.address import Address as Address
from braintree.amex_express_checkout_card import AmexExpressCheckoutCard as AmexExpressCheckoutCard
from braintree.android_pay_card import AndroidPayCard as AndroidPayCard
from braintree.apple_pay_card import ApplePayCard as ApplePayCard
from braintree.configuration import Configuration as Configuration
from braintree.credit_card import CreditCard as CreditCard
from braintree.error_result import ErrorResult as ErrorResult
from braintree.europe_bank_account import EuropeBankAccount as EuropeBankAccount
from braintree.exceptions.not_found_error import NotFoundError as NotFoundError
from braintree.ids_search import IdsSearch as IdsSearch
from braintree.masterpass_card import MasterpassCard as MasterpassCard
from braintree.paypal_account import PayPalAccount as PayPalAccount
from braintree.resource import Resource as Resource
from braintree.resource_collection import ResourceCollection as ResourceCollection
from braintree.samsung_pay_card import SamsungPayCard as SamsungPayCard
from braintree.successful_result import SuccessfulResult as SuccessfulResult
from braintree.us_bank_account import UsBankAccount as UsBankAccount
from braintree.util.http import Http as Http
from braintree.venmo_account import VenmoAccount as VenmoAccount
from braintree.visa_checkout_card import VisaCheckoutCard as VisaCheckoutCard

class Customer(Resource):
    @staticmethod
    def all(): ...
    @staticmethod
    def create(params: Any | None = ...): ...
    @staticmethod
    def delete(customer_id): ...
    @staticmethod
    def find(customer_id, association_filter_id: Any | None = ...): ...
    @staticmethod
    def search(*query): ...
    @staticmethod
    def update(customer_id, params: Any | None = ...): ...
    @staticmethod
    def create_signature(): ...
    @staticmethod
    def update_signature(): ...
    payment_methods: Any
    credit_cards: Any
    addresses: Any
    paypal_accounts: Any
    apple_pay_cards: Any
    android_pay_cards: Any
    amex_express_checkout_cards: Any
    europe_bank_accounts: Any
    venmo_accounts: Any
    us_bank_accounts: Any
    visa_checkout_cards: Any
    masterpass_cards: Any
    samsung_pay_cards: Any
    def __init__(self, gateway, attributes) -> None: ...



from _typeshed import Self
from collections.abc import Callable, Iterator
from logging import Logger
from typing import IO
from typing_extensions import TypeAlias

from paramiko.channel import Channel
from paramiko.sftp import BaseSFTP
from paramiko.sftp_attr import SFTPAttributes
from paramiko.sftp_file import SFTPFile
from paramiko.transport import Transport
from paramiko.util import ClosingContextManager

_Callback: TypeAlias = Callable[[int, int], object]

b_slash: bytes

class SFTPClient(BaseSFTP, ClosingContextManager):
    sock: Channel
    ultra_debug: bool
    request_number: int
    logger: Logger
    def __init__(self, sock: Channel) -> None: ...
    @classmethod
    def from_transport(
        cls: type[Self], t: Transport, window_size: int | None = ..., max_packet_size: int | None = ...
    ) -> Self | None: ...
    def close(self) -> None: ...
    def get_channel(self) -> Channel | None: ...
    def listdir(self, path: str = ...) -> list[str]: ...
    def listdir_attr(self, path: str = ...) -> list[SFTPAttributes]: ...
    def listdir_iter(self, path: bytes | str = ..., read_aheads: int = ...) -> Iterator[SFTPAttributes]: ...
    def open(self, filename: bytes | str, mode: str = ..., bufsize: int = ...) -> SFTPFile: ...
    file = open
    def remove(self, path: bytes | str) -> None: ...
    unlink = remove
    def rename(self, oldpath: bytes | str, newpath: bytes | str) -> None: ...
    def posix_rename(self, oldpath: bytes | str, newpath: bytes | str) -> None: ...
    def mkdir(self, path: bytes | str, mode: int = ...) -> None: ...
    def rmdir(self, path: bytes | str) -> None: ...
    def stat(self, path: bytes | str) -> SFTPAttributes: ...
    def lstat(self, path: bytes | str) -> SFTPAttributes: ...
    def symlink(self, source: bytes | str, dest: bytes | str) -> None: ...
    def chmod(self, path: bytes | str, mode: int) -> None: ...
    def chown(self, path: bytes | str, uid: int, gid: int) -> None: ...
    def utime(self, path: bytes | str, times: tuple[float, float] | None) -> None: ...
    def truncate(self, path: bytes | str, size: int) -> None: ...
    def readlink(self, path: bytes | str) -> str | None: ...
    def normalize(self, path: bytes | str) -> str: ...
    def chdir(self, path: None | bytes | str = ...) -> None: ...
    def getcwd(self) -> str | None: ...
    def putfo(
        self, fl: IO[bytes], remotepath: bytes | str, file_size: int = ..., callback: _Callback | None = ..., confirm: bool = ...
    ) -> SFTPAttributes: ...
    def put(
        self, localpath: bytes | str, remotepath: bytes | str, callback: _Callback | None = ..., confirm: bool = ...
    ) -> SFTPAttributes: ...
    def getfo(self, remotepath: bytes | str, fl: IO[bytes], callback: _Callback | None = ..., prefetch: bool = ...) -> int: ...
    def get(
        self, remotepath: bytes | str, localpath: bytes | str, callback: _Callback | None = ..., prefetch: bool = ...
    ) -> None: ...

class SFTP(SFTPClient): ...



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class PictureLocking(Serialisable):
    tagname: str
    namespace: Any
    noCrop: Any
    noGrp: Any
    noSelect: Any
    noRot: Any
    noChangeAspect: Any
    noMove: Any
    noResize: Any
    noEditPoints: Any
    noAdjustHandles: Any
    noChangeArrowheads: Any
    noChangeShapeType: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        noCrop: Any | None = ...,
        noGrp: Any | None = ...,
        noSelect: Any | None = ...,
        noRot: Any | None = ...,
        noChangeAspect: Any | None = ...,
        noMove: Any | None = ...,
        noResize: Any | None = ...,
        noEditPoints: Any | None = ...,
        noAdjustHandles: Any | None = ...,
        noChangeArrowheads: Any | None = ...,
        noChangeShapeType: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...

class NonVisualPictureProperties(Serialisable):
    tagname: str
    preferRelativeResize: Any
    picLocks: Any
    extLst: Any
    __elements__: Any
    def __init__(self, preferRelativeResize: Any | None = ..., picLocks: Any | None = ..., extLst: Any | None = ...) -> None: ...

class PictureNonVisual(Serialisable):
    tagname: str
    cNvPr: Any
    cNvPicPr: Any
    __elements__: Any
    def __init__(self, cNvPr: Any | None = ..., cNvPicPr: Any | None = ...) -> None: ...

class PictureFrame(Serialisable):
    tagname: str
    macro: Any
    fPublished: Any
    nvPicPr: Any
    blipFill: Any
    spPr: Any
    graphicalProperties: Any
    style: Any
    __elements__: Any
    def __init__(
        self,
        macro: Any | None = ...,
        fPublished: Any | None = ...,
        nvPicPr: Any | None = ...,
        blipFill: Any | None = ...,
        spPr: Any | None = ...,
        style: Any | None = ...,
    ) -> None: ...



from typing import Any

class DependencyProcessor:
    prop: Any
    cascade: Any
    mapper: Any
    parent: Any
    secondary: Any
    direction: Any
    post_update: Any
    passive_deletes: Any
    passive_updates: Any
    enable_typechecks: Any
    sort_key: Any
    key: Any
    def __init__(self, prop) -> None: ...
    @classmethod
    def from_relationship(cls, prop): ...
    def hasparent(self, state): ...
    def per_property_preprocessors(self, uow) -> None: ...
    def per_property_flush_actions(self, uow) -> None: ...
    def per_state_flush_actions(self, uow, states, isdelete) -> None: ...
    def presort_deletes(self, uowcommit, states): ...
    def presort_saves(self, uowcommit, states): ...
    def process_deletes(self, uowcommit, states) -> None: ...
    def process_saves(self, uowcommit, states) -> None: ...
    def prop_has_changes(self, uowcommit, states, isdelete): ...

class OneToManyDP(DependencyProcessor):
    def per_property_dependencies(
        self, uow, parent_saves, child_saves, parent_deletes, child_deletes, after_save, before_delete
    ) -> None: ...
    def per_state_dependencies(
        self, uow, save_parent, delete_parent, child_action, after_save, before_delete, isdelete, childisdelete
    ) -> None: ...
    def presort_deletes(self, uowcommit, states) -> None: ...
    def presort_saves(self, uowcommit, states) -> None: ...
    def process_deletes(self, uowcommit, states) -> None: ...
    def process_saves(self, uowcommit, states) -> None: ...

class ManyToOneDP(DependencyProcessor):
    def __init__(self, prop) -> None: ...
    def per_property_dependencies(
        self, uow, parent_saves, child_saves, parent_deletes, child_deletes, after_save, before_delete
    ) -> None: ...
    def per_state_dependencies(
        self, uow, save_parent, delete_parent, child_action, after_save, before_delete, isdelete, childisdelete
    ) -> None: ...
    def presort_deletes(self, uowcommit, states) -> None: ...
    def presort_saves(self, uowcommit, states) -> None: ...
    def process_deletes(self, uowcommit, states) -> None: ...
    def process_saves(self, uowcommit, states) -> None: ...

class DetectKeySwitch(DependencyProcessor):
    def per_property_preprocessors(self, uow) -> None: ...
    def per_property_flush_actions(self, uow) -> None: ...
    def per_state_flush_actions(self, uow, states, isdelete) -> None: ...
    def presort_deletes(self, uowcommit, states) -> None: ...
    def presort_saves(self, uow, states) -> None: ...
    def prop_has_changes(self, uow, states, isdelete): ...
    def process_deletes(self, uowcommit, states) -> None: ...
    def process_saves(self, uowcommit, states) -> None: ...

class ManyToManyDP(DependencyProcessor):
    def per_property_dependencies(
        self, uow, parent_saves, child_saves, parent_deletes, child_deletes, after_save, before_delete
    ) -> None: ...
    def per_state_dependencies(
        self, uow, save_parent, delete_parent, child_action, after_save, before_delete, isdelete, childisdelete
    ) -> None: ...
    def presort_deletes(self, uowcommit, states) -> None: ...
    def presort_saves(self, uowcommit, states) -> None: ...
    def process_deletes(self, uowcommit, states) -> None: ...
    def process_saves(self, uowcommit, states) -> None: ...



from _typeshed import Incomplete, SupportsWrite
from collections.abc import Iterable, Mapping
from typing import Generic, NoReturn, TypeVar, overload

from ..auto import tqdm as tqdm_auto
from .utils_worker import MonoWorker

__all__ = ["TelegramIO", "tqdm_telegram", "ttgrange", "tqdm", "trange"]

class TelegramIO(MonoWorker):
    API: str
    token: Incomplete
    chat_id: Incomplete
    session: Incomplete
    text: Incomplete
    def __init__(self, token, chat_id) -> None: ...
    @property
    def message_id(self): ...
    def write(self, s: str) -> Incomplete | None: ...
    def delete(self): ...

_T = TypeVar("_T")

class tqdm_telegram(Generic[_T], tqdm_auto[_T]):
    tgio: Incomplete
    @overload
    def __init__(
        self,
        iterable: Iterable[_T],
        desc: str | None = ...,
        total: float | None = ...,
        leave: bool | None = ...,
        file: SupportsWrite[str] | None = ...,
        ncols: int | None = ...,
        mininterval: float = ...,
        maxinterval: float = ...,
        miniters: float | None = ...,
        ascii: bool | str | None = ...,
        disable: bool = ...,
        unit: str = ...,
        unit_scale: bool | float = ...,
        dynamic_ncols: bool = ...,
        smoothing: float = ...,
        bar_format: str | None = ...,
        initial: float = ...,
        position: int | None = ...,
        postfix: Mapping[str, object] | str | None = ...,
        unit_divisor: float = ...,
        write_bytes: bool | None = ...,
        lock_args: tuple[bool | None, float | None] | tuple[bool | None] | None = ...,
        nrows: int | None = ...,
        colour: str | None = ...,
        delay: float | None = ...,
        gui: bool = ...,
        token: str = ...,
        chat_id: str = ...,
        **kwargs,
    ) -> None: ...
    @overload
    def __init__(
        self: tqdm_telegram[NoReturn],
        iterable: None = ...,
        desc: str | None = ...,
        total: float | None = ...,
        leave: bool | None = ...,
        file: SupportsWrite[str] | None = ...,
        ncols: int | None = ...,
        mininterval: float = ...,
        maxinterval: float = ...,
        miniters: float | None = ...,
        ascii: bool | str | None = ...,
        disable: bool = ...,
        unit: str = ...,
        unit_scale: bool | float = ...,
        dynamic_ncols: bool = ...,
        smoothing: float = ...,
        bar_format: str | None = ...,
        initial: float = ...,
        position: int | None = ...,
        postfix: Mapping[str, object] | str | None = ...,
        unit_divisor: float = ...,
        write_bytes: bool | None = ...,
        lock_args: tuple[bool | None, float | None] | tuple[bool | None] | None = ...,
        nrows: int | None = ...,
        colour: str | None = ...,
        delay: float | None = ...,
        gui: bool = ...,
        token: str = ...,
        chat_id: str = ...,
        **kwargs,
    ) -> None: ...
    def display(self, *, msg: str | None = ..., pos: int | None = ..., close: bool = ..., bar_style: Incomplete = ..., check_delay: bool = ...) -> None: ...  # type: ignore[override]
    def clear(self, *args, **kwargs) -> None: ...
    def close(self) -> None: ...

def ttgrange(*args, **kwargs) -> tqdm_telegram[int]: ...

tqdm = tqdm_telegram
trange = ttgrange



from collections.abc import Iterator

class ChevronError(SyntaxError): ...

def grab_literal(template: str, l_del: str | None) -> tuple[str, str]: ...  # undocumented
def l_sa_check(template: str, literal: str, is_standalone: bool) -> bool | None: ...  # undocumented
def r_sa_check(template: str, tag_type: str, is_standalone: bool) -> bool: ...  # undocumented
def parse_tag(template: str, l_del: str | None, r_del: str | None) -> tuple[tuple[str, str], str]: ...  # undocumented
def tokenize(
    template: str, def_ldel: str | None = ..., def_rdel: str | None = ...
) -> Iterator[tuple[str, str]]: ...  # undocumented



from typing import Any
from typing_extensions import TypedDict

from . import MQTTException as MQTTException

class _SubscribeOptionsJson(TypedDict):
    QoS: int
    noLocal: bool
    retainAsPublished: bool
    retainHandling: int

class SubscribeOptions:
    RETAIN_SEND_ON_SUBSCRIBE: int
    RETAIN_SEND_IF_NEW_SUB: int
    RETAIN_DO_NOT_SEND: int
    names: list[str]
    Qos: int
    noLocal: bool
    retainAsPublished: bool
    retainHandling: int
    def __init__(self, qos: int = ..., noLocal: bool = ..., retainAsPublished: bool = ..., retainHandling: int = ...) -> None: ...
    def __setattr__(self, name: str, value: Any) -> None: ...
    def pack(self) -> bytes: ...
    def unpack(self, buffer: bytes | bytearray) -> int: ...
    def json(self) -> _SubscribeOptionsJson: ...



from _typeshed import StrOrBytesPath, SupportsWrite
from collections.abc import Iterable, Mapping
from distutils.cmd import Command
from typing import IO, Any

class DistributionMetadata:
    def __init__(self, path: int | StrOrBytesPath | None = ...) -> None: ...
    name: str | None
    version: str | None
    author: str | None
    author_email: str | None
    maintainer: str | None
    maintainer_email: str | None
    url: str | None
    license: str | None
    description: str | None
    long_description: str | None
    keywords: str | list[str] | None
    platforms: str | list[str] | None
    classifiers: str | list[str] | None
    download_url: str | None
    provides: list[str] | None
    requires: list[str] | None
    obsoletes: list[str] | None
    def read_pkg_file(self, file: IO[str]) -> None: ...
    def write_pkg_info(self, base_dir: str) -> None: ...
    def write_pkg_file(self, file: SupportsWrite[str]) -> None: ...
    def get_name(self) -> str: ...
    def get_version(self) -> str: ...
    def get_fullname(self) -> str: ...
    def get_author(self) -> str: ...
    def get_author_email(self) -> str: ...
    def get_maintainer(self) -> str: ...
    def get_maintainer_email(self) -> str: ...
    def get_contact(self) -> str: ...
    def get_contact_email(self) -> str: ...
    def get_url(self) -> str: ...
    def get_license(self) -> str: ...
    def get_licence(self) -> str: ...
    def get_description(self) -> str: ...
    def get_long_description(self) -> str: ...
    def get_keywords(self) -> str | list[str]: ...
    def get_platforms(self) -> str | list[str]: ...
    def get_classifiers(self) -> str | list[str]: ...
    def get_download_url(self) -> str: ...
    def get_requires(self) -> list[str]: ...
    def set_requires(self, value: Iterable[str]) -> None: ...
    def get_provides(self) -> list[str]: ...
    def set_provides(self, value: Iterable[str]) -> None: ...
    def get_obsoletes(self) -> list[str]: ...
    def set_obsoletes(self, value: Iterable[str]) -> None: ...

class Distribution:
    cmdclass: dict[str, type[Command]]
    metadata: DistributionMetadata
    def __init__(self, attrs: Mapping[str, Any] | None = ...) -> None: ...
    def get_option_dict(self, command: str) -> dict[str, tuple[str, str]]: ...
    def parse_config_files(self, filenames: Iterable[str] | None = ...) -> None: ...
    def get_command_obj(self, command: str, create: bool = ...) -> Command | None: ...



from io import TextIOWrapper
from typing import Any

PY2: bool
PY3: bool
WIN: bool
string_types: tuple[str]
integer_types: tuple[int]
class_types: tuple[type]

def unquote_bytes_to_wsgi(bytestring: bytes) -> str: ...
def text_(s: str, encoding: str = ..., errors: str = ...) -> str: ...
def tostr(s: str) -> str: ...
def tobytes(s: str) -> bytes: ...

exec_: Any

def reraise(tp: Any, value: BaseException, tb: str | None = ...) -> None: ...

MAXINT: int
HAS_IPV6: bool
IPPROTO_IPV6: int
IPV6_V6ONLY: int

def set_nonblocking(fd: TextIOWrapper) -> None: ...

ResourceWarning: Warning

def qualname(cls: Any) -> str: ...



from typing import Any

Marker: Any
listElementsMap: Any

class Node:
    name: Any
    parent: Any
    value: Any
    attributes: Any
    childNodes: Any
    def __init__(self, name) -> None: ...
    def appendChild(self, node) -> None: ...
    def insertText(self, data, insertBefore: Any | None = ...) -> None: ...
    def insertBefore(self, node, refNode) -> None: ...
    def removeChild(self, node) -> None: ...
    def reparentChildren(self, newParent) -> None: ...
    def cloneNode(self) -> None: ...
    def hasContent(self) -> None: ...

class ActiveFormattingElements(list[Any]):
    def append(self, node) -> None: ...
    def nodesEqual(self, node1, node2): ...

class TreeBuilder:
    documentClass: Any
    elementClass: Any
    commentClass: Any
    doctypeClass: Any
    fragmentClass: Any
    defaultNamespace: str
    def __init__(self, namespaceHTMLElements) -> None: ...
    openElements: Any
    activeFormattingElements: Any
    headPointer: Any
    formPointer: Any
    insertFromTable: bool
    document: Any
    def reset(self) -> None: ...
    def elementInScope(self, target, variant: Any | None = ...): ...
    def reconstructActiveFormattingElements(self) -> None: ...
    def clearActiveFormattingElements(self) -> None: ...
    def elementInActiveFormattingElements(self, name): ...
    def insertRoot(self, token) -> None: ...
    def insertDoctype(self, token) -> None: ...
    def insertComment(self, token, parent: Any | None = ...) -> None: ...
    def createElement(self, token): ...
    def insertElementNormal(self, token): ...
    def insertElementTable(self, token): ...
    def insertText(self, data, parent: Any | None = ...) -> None: ...
    def getTableMisnestedNodePosition(self): ...
    def generateImpliedEndTags(self, exclude: Any | None = ...) -> None: ...
    def getDocument(self): ...
    def getFragment(self): ...
    def testSerializer(self, node) -> None: ...



# https://pyinstaller.org/en/stable/hooks.html#module-PyInstaller.compat
from _typeshed import FileDescriptor, GenericPath, StrOrBytesPath
from collections.abc import Iterable
from types import ModuleType
from typing import AnyStr, overload
from typing_extensions import Literal, TypeAlias

_OpenFile: TypeAlias = StrOrBytesPath | FileDescriptor

is_64bits: bool
is_py35: bool
is_py36: bool
is_py37: bool
is_py38: bool
is_py39: bool
is_py310: bool
is_py311: bool
is_win: bool
is_win_10: bool
is_win_wine: bool
is_cygwin: bool
is_darwin: bool
is_linux: bool
is_solar: bool
is_aix: bool
is_freebsd: bool
is_openbsd: bool
is_hpux: bool
is_unix: bool
is_musl: bool
is_macos_11_compat: bool
is_macos_11_native: bool
is_macos_11: bool
PYDYLIB_NAMES: set[str]
base_prefix: str
is_venv: bool
is_virtualenv: bool
is_conda: bool
is_pure_conda: bool
python_executable: str
is_ms_app_store: bool
BYTECODE_MAGIC: bytes
EXTENSION_SUFFIXES: list[str]
ALL_SUFFIXES: list[str]

architecture: Literal["64bit", "n32bit", "32bit"]
system: Literal["Cygwin", "Linux", "Darwin", "Java", "Windows"]
machine: Literal["sw_64", "loongarch64", "arm", "intel", "ppc", "mips", "riscv", "s390x", "unknown", None]

def is_wine_dll(filename: _OpenFile) -> bool: ...
@overload
def getenv(name: str, default: str) -> str: ...
@overload
def getenv(name: str, default: None = ...) -> str | None: ...
def setenv(name: str, value: str) -> None: ...
def unsetenv(name: str) -> None: ...
def exec_command(
    *cmdargs: str, encoding: str | None = ..., raise_enoent: bool | None = ..., **kwargs: int | bool | Iterable[int] | None
) -> str: ...
def exec_command_rc(*cmdargs: str, **kwargs: float | bool | Iterable[int] | None) -> int: ...
def exec_command_stdout(
    *command_args: str, encoding: str | None = ..., **kwargs: float | str | bytes | bool | Iterable[int] | None
) -> str: ...
def exec_command_all(
    *cmdargs: str, encoding: str | None = ..., **kwargs: int | bool | Iterable[int] | None
) -> tuple[int, str, str]: ...
def exec_python(*args: str, **kwargs: str | None) -> str: ...
def exec_python_rc(*args: str, **kwargs: str | None) -> int: ...
def expand_path(path: GenericPath[AnyStr]) -> AnyStr: ...
def getsitepackages(prefixes: Iterable[str] | None = ...) -> list[str]: ...
def importlib_load_source(name: str, pathname: str) -> ModuleType: ...

PY3_BASE_MODULES: set[str]
PURE_PYTHON_MODULE_TYPES: set[str]
SPECIAL_MODULE_TYPES: set[str]
BINARY_MODULE_TYPES: set[str]
VALID_MODULE_TYPES: set[str]
BAD_MODULE_TYPES: set[str]
ALL_MODULE_TYPES: set[str]
MODULE_TYPES_TO_TOC_DICT: dict[str, str]

def check_requirements() -> None: ...



from _typeshed import Incomplete
from typing import Any

# Necessary for mypy to not throw AssertionError with win32com.client
class CDispatch:
    def __init__(
        self,
        IDispatch,
        olerepr,
        userName: Incomplete | None = ...,
        UnicodeToString: Incomplete | None = ...,
        lazydata: Incomplete | None = ...,
    ) -> None: ...
    def __call__(self, *args): ...
    def __bool__(self): ...
    def __dir__(self): ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...
    def __int__(self): ...
    def __len__(self): ...
    def __getitem__(self, index): ...
    def __setitem__(self, index, *args): ...
    def __LazyMap__(self, attr): ...
    def __AttrToID__(self, attr): ...
    ob: Incomplete
    # CDispatch objects are dynamically generated and too complex to type
    def __getattr__(self, attr: str) -> Any: ...
    def __setattr__(self, attr: str, value: Any) -> None: ...
